# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Provide math calls that uses intrinsics or libc math functions.

The bitcast/comparison helpers and the @lower codegen for the math module
(isnan/isinf/isfinite, copysign, frexp/ldexp, atan2/hypot, radians/degrees,
pow/nextafter/gcd, ...) built llvmlite IR and were filtered out on the MLIR
path (math is lowered by numba_cuda_mlir.lowering.math). They are gone. What
remains is pure Python: the float min/max constants (still used by
cmathimpl) and the ``_unsigned`` typing overload used by integer helpers.
"""

import numpy as np

from numba_cuda_mlir.numba_cuda.core.imputils import Registry
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.extending import overload

registry = Registry("mathimpl")

# Helpers, shared with cmathimpl.
_NP_FLT_FINFO = np.finfo(np.dtype("float32"))
FLT_MAX = _NP_FLT_FINFO.max
FLT_MIN = _NP_FLT_FINFO.tiny

_NP_DBL_FINFO = np.finfo(np.dtype("float64"))
DBL_MAX = _NP_DBL_FINFO.max
DBL_MIN = _NP_DBL_FINFO.tiny

FLOAT_ABS_MASK = 0x7FFFFFFF
FLOAT_SIGN_MASK = 0x80000000
DOUBLE_ABS_MASK = 0x7FFFFFFFFFFFFFFF
DOUBLE_SIGN_MASK = 0x8000000000000000


def _unsigned(T):
    """Convert integer to unsigned integer of equivalent width."""
    pass


@overload(_unsigned)
def _unsigned_impl(T):
    if T in types.unsigned_domain:
        return lambda T: T
    elif T in types.signed_domain:
        newT = getattr(types, "uint{}".format(T.bitwidth))
        return lambda T: newT(T)



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Optional, Sequence

from ...ir import (
    Type,
    Value,
    Operation,
    OpView,
    Attribute,
    ArrayAttr,
    StringAttr,
    F64Type,
    IntegerType,
    IntegerAttr,
    FloatAttr,
    BoolAttr,
)
from .._transform_tune_extension_ops_gen import *
from .._transform_tune_extension_ops_gen import _Dialect

try:
    from .._ods_common import (
        get_op_result_or_value as _get_op_result_or_value,
        _cext as _ods_cext,
    )
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Union


@_ods_cext.register_operation(_Dialect, replace=True)
class KnobOp(KnobOp):
    def __init__(
        self,
        result: Type,  # !transform.any_param or !transform.param<Type>
        name: Union[StringAttr, str],
        options: Union[
            ArrayAttr, Sequence[Union[Attribute, bool, int, float, str]], Attribute
        ],
        *,
        selected: Optional[Union[Attribute, bool, int, float, str]] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(name, str):
            name = StringAttr.get(name)

        def map_to_attr(value):
            if isinstance(value, bool):
                return BoolAttr.get(value)
            if isinstance(value, int):
                return IntegerAttr.get(IntegerType.get_signless(64), value)
            if isinstance(value, float):
                return FloatAttr.get(F64Type.get(), value)
            if isinstance(value, str):
                return StringAttr.get(value)
            assert isinstance(value, Attribute)
            return value

        if isinstance(options, Sequence) and not isinstance(options, ArrayAttr):
            options = ArrayAttr.get([map_to_attr(opt) for opt in options])

        super().__init__(
            result,
            name,
            options,
            selected=selected and map_to_attr(selected),
            loc=loc,
            ip=ip,
        )


def knob(
    result: Type,  # !transform.any_param or !transform.param<Type>
    name: Union[StringAttr, str],
    options: Union[
        ArrayAttr, Sequence[Union[Attribute, bool, int, float, str]], Attribute
    ],
    *,
    selected: Optional[Union[Attribute, bool, int, float, str]] = None,
    loc=None,
    ip=None,
):
    return KnobOp(result, name, options, selected=selected, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class AlternativesOp(AlternativesOp):
    def __init__(
        self,
        results: Sequence[Type],
        name: Union[StringAttr, str],
        num_alternatives: int,
        *,
        selected_region: Optional[
            Union[int, IntegerAttr, Value, Operation, OpView]
        ] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(name, str):
            name = StringAttr.get(name)

        selected_region_attr = selected_region_param = None
        if isinstance(selected_region, IntegerAttr):
            selected_region_attr = selected_region
        elif isinstance(selected_region, int):
            selected_region_attr = IntegerAttr.get(
                IntegerType.get_signless(32), selected_region
            )
        elif isinstance(selected_region, (Value, Operation, OpView)):
            selected_region_param = _get_op_result_or_value(selected_region)

        super().__init__(
            results,
            name,
            num_alternatives,
            selected_region_attr=selected_region_attr,
            selected_region_param=selected_region_param,
            loc=loc,
            ip=ip,
        )
        for region in self.regions:
            region.blocks.append()


def alternatives(
    results: Sequence[Type],
    name: Union[StringAttr, str],
    num_alternatives: int,
    *,
    selected_region: Optional[Union[int, IntegerAttr, Value, Operation, OpView]] = None,
    loc=None,
    ip=None,
):
    return AlternativesOp(
        results, name, num_alternatives, selected_region=selected_region, loc=loc, ip=ip
    )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import operator

from numba_cuda_mlir import types
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.dialects import arith, scf
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir.mlir.dialect_exts import llvm

from numba_cuda_mlir.extending import (
    overload,
    overload_method,
    register_jitable,
    typing_registry as extending_typing_registry,
)
from numba_cuda_mlir.lowering_registry import LoweringRegistry
from numba_cuda_mlir.lowering_utilities import GEP_DYNAMIC_INDEX, true, false

from numba_cuda_mlir.numba_cuda.cpython.unicode import (
    _malloc_string,
    _get_code_point,
    _kind_to_byte_width,
    set_uint8,
    set_uint16,
    set_uint32,
    deref_uint8,
    deref_uint16,
    deref_uint32,
)

from numba_cuda_mlir.numba_cuda.core.pythonapi import (
    PY_UNICODE_1BYTE_KIND,
    PY_UNICODE_2BYTE_KIND,
    PY_UNICODE_4BYTE_KIND,
)

from numba_cuda_mlir.numba_cuda.core import errors

import numpy as np

registry = LoweringRegistry()
lower = registry.lower


def _unicode_eq_lower(builder, target, args, kwargs):
    """Lower unicode_type == unicode_type / StringLiteral.

    Constant-folds when both values are known Python strings at compile time.
    Otherwise compares length first, then byte-by-byte (ASCII/kind=1).
    Uses select + for loop (no scf.if with results needed).
    """
    from numba_cuda_mlir.lowering_utilities.string import (
        materialize_string_constant_if_needed,
    )

    lhs_raw = builder.load_var(args[0])
    rhs_raw = builder.load_var(args[1])

    if isinstance(lhs_raw, str) and isinstance(rhs_raw, str):
        result = arith.constant(result=T.bool(), value=(lhs_raw == rhs_raw))
        builder.store_var(target, result)
        return

    lhs_val = materialize_string_constant_if_needed(builder.mlir_gpu_module, lhs_raw)
    rhs_val = materialize_string_constant_if_needed(builder.mlir_gpu_module, rhs_raw)

    # Extract lengths (field 1)
    lhs_len = llvm.extractvalue(T.i64(), lhs_val, [1])
    rhs_len = llvm.extractvalue(T.i64(), rhs_val, [1])

    # Extract data pointers (field 0)
    lhs_data = llvm.extractvalue(llvm.ptr(), lhs_val, [0])
    rhs_data = llvm.extractvalue(llvm.ptr(), rhs_val, [0])

    # Compare lengths
    len_eq = arith.cmpi(arith.CmpIPredicate.eq, lhs_len, rhs_len)

    # Compare bytes in a loop: accumulate mismatch count.
    # We compare raw bytes (works for kind=1 ASCII strings).
    zero_i64 = arith.constant(T.i64(), 0)
    one_i64 = arith.constant(T.i64(), 1)
    zero_mismatches = arith.constant(T.i64(), 0)

    # scf.for with iter_arg: count mismatches
    loop = scf.ForOp(zero_i64, lhs_len, one_i64, [zero_mismatches])
    iv = loop.body.arguments[0]
    mismatch_count = loop.body.arguments[1]
    with ir.InsertionPoint(loop.body):
        lhs_byte_ptr = llvm.getelementptr(
            llvm.ptr(), lhs_data, [iv], [GEP_DYNAMIC_INDEX], T.i8(), None
        )
        rhs_byte_ptr = llvm.getelementptr(
            llvm.ptr(), rhs_data, [iv], [GEP_DYNAMIC_INDEX], T.i8(), None
        )
        lhs_byte = llvm.load(T.i8(), lhs_byte_ptr)
        rhs_byte = llvm.load(T.i8(), rhs_byte_ptr)
        bytes_ne = arith.cmpi(arith.CmpIPredicate.ne, lhs_byte, rhs_byte)
        ne_as_i64 = arith.extui(T.i64(), bytes_ne)
        new_count = arith.addi(mismatch_count, ne_as_i64)
        scf.YieldOp([new_count])

    # Result: lengths equal AND no mismatches
    no_mismatches = arith.cmpi(arith.CmpIPredicate.eq, loop.results[0], zero_i64)
    result = arith.andi(len_eq, no_mismatches)
    builder.store_var(target, result)


# Register for all combinations of UnicodeType and StringLiteral
@lower(operator.eq, types.UnicodeType, types.UnicodeType)
@lower(operator.eq, types.UnicodeType, types.StringLiteral)
@lower(operator.eq, types.StringLiteral, types.UnicodeType)
def _lower_eq(builder, target, args, kwargs):
    _unicode_eq_lower(builder, target, args, kwargs)


@lower(operator.ne, types.UnicodeType, types.UnicodeType)
@lower(operator.ne, types.UnicodeType, types.StringLiteral)
@lower(operator.ne, types.StringLiteral, types.UnicodeType)
def _lower_ne(builder, target, args, kwargs):
    _unicode_eq_lower(builder, target, args, kwargs)
    eq_result = builder.load_var(target)
    ne_result = arith.xori(eq_result, true())
    builder.store_var(target, ne_result)


@register_jitable(typing_registry=extending_typing_registry)
def _empty_string_numba_cuda_mlir(kind, length, is_ascii=0):
    char_width = _kind_to_byte_width(kind)
    s = _malloc_string(kind, char_width, length, is_ascii)
    _set_code_point(s, length, np.uint32(0))
    return s


@register_jitable(typing_registry=extending_typing_registry)
def _ascii_upper_numba_cuda_mlir(data, res):
    for idx in range(len(data)):
        ch = _get_code_point(data, idx)
        if 97 <= ch <= 122:  # ord('a') .. ord('z')
            ch = ch - 32
        _set_code_point(res, idx, ch)


@register_jitable(typing_registry=extending_typing_registry)
def _ascii_lower_numba_cuda_mlir(data, res):
    for idx in range(len(data)):
        ch = _get_code_point(data, idx)
        if 65 <= ch <= 90:  # ord('A') .. ord('Z')
            ch = ch + 32
        _set_code_point(res, idx, ch)


@overload_method(types.UnicodeType, "upper", typing_registry=extending_typing_registry)
def unicode_upper(data):
    def impl(data):
        length = len(data)
        if length == 0:
            return _empty_string_numba_cuda_mlir(data._kind, length, data._is_ascii)
        res = _empty_string_numba_cuda_mlir(data._kind, length, 1)
        _ascii_upper_numba_cuda_mlir(data, res)
        return res

    return impl


@overload_method(types.UnicodeType, "lower", typing_registry=extending_typing_registry)
def unicode_lower(data):
    def impl(data):
        length = len(data)
        if length == 0:
            return _empty_string_numba_cuda_mlir(data._kind, length, data._is_ascii)
        res = _empty_string_numba_cuda_mlir(data._kind, length, 1)
        _ascii_lower_numba_cuda_mlir(data, res)
        return res

    return impl


@overload(len, typing_registry=extending_typing_registry)
def unicode_len(s):
    if isinstance(s, types.UnicodeType):

        def len_impl(s):
            return s._length

        return len_impl


@register_jitable(_nrt=False, typing_registry=extending_typing_registry)
def _set_code_point(a, i, ch):
    # WARNING: This method is very dangerous:
    #   * Assumes that data contents can be changed (only allowed for new
    #     strings)
    #   * Assumes that the kind of unicode string is sufficiently wide to
    #     accept ch.  Will truncate ch to make it fit.
    #   * Assumes that i is within the valid boundaries of the function
    if a._kind == PY_UNICODE_1BYTE_KIND:
        set_uint8(a._data, i, ch)
    elif a._kind == PY_UNICODE_2BYTE_KIND:
        set_uint16(a._data, i, ch)
    elif a._kind == PY_UNICODE_4BYTE_KIND:
        set_uint32(a._data, i, ch)
    else:
        raise AssertionError("Unexpected unicode representation in _set_code_point")


# Disable RefCt for performance.
@register_jitable(_nrt=False, typing_registry=extending_typing_registry)
def _get_code_point(a, i):
    if a._kind == PY_UNICODE_1BYTE_KIND:
        return deref_uint8(a._data, i)
    elif a._kind == PY_UNICODE_2BYTE_KIND:
        return deref_uint16(a._data, i)
    elif a._kind == PY_UNICODE_4BYTE_KIND:
        return deref_uint32(a._data, i)
    else:
        # there's also a wchar kind, but that's one of the above,
        # so skipping for this example
        return 0



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import linecache
import os
from typing import NamedTuple

from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir.lowering_utilities import context as numba_cuda_mlir_context
from numba_cuda_mlir.numba_cuda.datamodel.models import ArrayModel
from numba_cuda_mlir.numba_cuda.types.ext_types import Bfloat16, GridGroup


class _DILocalVarInfo(NamedTuple):
    name: str
    line: int
    type_str: str
    arg_index: int | None


_BYTE_SIZE_BITS = 8
_BFLOAT16_BITS = 16
_GRID_GROUP_BITS = 64
_INT_LITERAL_BITS = 64
_POINTER_BITS = 64


def _mlir_string_attr(value):
    value = str(value).replace("\\", "/").replace('"', '\\"')
    return f'"{value}"'


def _basic_di_type(name, bits, encoding):
    return (
        f"#llvm.di_basic_type<tag = DW_TAG_base_type, name = {_mlir_string_attr(name)}, "
        f"sizeInBits = {bits}, encoding = DW_ATE_{encoding}>"
    )


def _cpointer_di_type(numba_type):
    pointee_di = _numba_type_to_di_type_str(numba_type.dtype)
    if pointee_di is None:
        pointee_di = _basic_di_type("byte", _BYTE_SIZE_BITS, "unsigned_char")
    return _derived_di_type(
        "DW_TAG_pointer_type",
        base_type=pointee_di,
        size_bits=_POINTER_BITS,
    )


def _enum_member_di_type(numba_type):
    dtype = numba_type.dtype
    if not isinstance(dtype, types.Integer):
        return None
    encoding = "signed" if dtype.signed else "unsigned"
    # NOTE: The current MLIR toolchain does not provide #llvm.di_enumerator and
    # lowering rejects DW_TAG_enumerator encoded via #llvm.di_derived_type in
    # the modern LLVM translation path. Until #llvm.di_enumerator is available,
    # emit a stable scalar DI type instead of invalid enum nodes.
    # Keep str(numba_type) by design.
    return _basic_di_type(str(numba_type), dtype.bitwidth, encoding)


def _complex_di_type(name, float_bits):
    """Build a DICompositeType string for a complex number."""
    float_type = _basic_di_type(f"float{float_bits}", float_bits, "float")
    real_member = (
        f"#llvm.di_derived_type<tag = DW_TAG_member, "
        f"name = {_mlir_string_attr('real')}, baseType = {float_type}, "
        f"sizeInBits = {float_bits}, offsetInBits = 0>"
    )
    imag_member = (
        f"#llvm.di_derived_type<tag = DW_TAG_member, "
        f"name = {_mlir_string_attr('imag')}, baseType = {float_type}, "
        f"sizeInBits = {float_bits}, offsetInBits = {float_bits}>"
    )
    return (
        f"#llvm.di_composite_type<tag = DW_TAG_structure_type, "
        f"name = {_mlir_string_attr(name)}, sizeInBits = {float_bits * 2}, "
        f"elements = {real_member}, {imag_member}>"
    )


def _derived_di_type(
    tag,
    *,
    name=None,
    base_type=None,
    size_bits=None,
    offset_bits=None,
    flags=None,
    extra_data=None,
):
    params = [f"tag = {tag}"]
    if name is not None:
        params.append(f"name = {_mlir_string_attr(name)}")
    if base_type is not None:
        params.append(f"baseType = {base_type}")
    if size_bits is not None:
        params.append(f"sizeInBits = {size_bits}")
    if offset_bits is not None:
        params.append(f"offsetInBits = {offset_bits}")
    if flags is not None:
        params.append(f"flags = {flags}")
    if extra_data is not None:
        params.append(f"extraData = {extra_data} : i8")
    return f"#llvm.di_derived_type<{', '.join(params)}>"


def _composite_di_type(
    tag,
    *,
    name=None,
    base_type=None,
    size_bits=None,
    elements=None,
    identifier=None,
    discriminator=None,
):
    params = [f"tag = {tag}"]
    if name is not None:
        params.append(f"name = {_mlir_string_attr(name)}")
    if base_type is not None:
        params.append(f"baseType = {base_type}")
    if size_bits is not None:
        params.append(f"sizeInBits = {size_bits}")
    if identifier is not None:
        params.append(f"identifier = {_mlir_string_attr(identifier)}")
    if discriminator is not None:
        params.append(f"discriminator = {discriminator}")
    if elements:
        params.append(f"elements = {', '.join(elements)}")
    return f"#llvm.di_composite_type<{', '.join(params)}>"


def _subrange_di_type(count):
    return f"#llvm.di_subrange<count = {count}>"


def _llvm_type_str(numba_type):
    numba_type = _strip_literal_type(numba_type)
    match numba_type:
        case types.Boolean():
            return "i8"
        case types.Integer(bitwidth=bw):
            return f"i{bw}"
        case types.Float(bitwidth=16):
            return "half"
        case types.Float(bitwidth=32):
            return "float"
        case types.Float(bitwidth=64):
            return "double"
        case types.UniTuple(dtype=dtype, count=count):
            elem_str = _llvm_type_str(dtype)
            return None if elem_str is None else f"[{count} x {elem_str}]"
        case types.BaseTuple():
            elem_strs = [_llvm_type_str(t) for t in numba_type.types]
            if any(elem_str is None for elem_str in elem_strs):
                return None
            return "{" + ", ".join(elem_strs) + "}"
        case _:
            return None


def _align_to_bits(offset_bits, alignment_bits):
    return (
        offset_bits
        if offset_bits % alignment_bits == 0
        else offset_bits + alignment_bits - offset_bits % alignment_bits
    )


def _type_alignment_bits(numba_type):
    numba_type = _strip_literal_type(numba_type)
    match numba_type:
        case types.Boolean():
            return _BYTE_SIZE_BITS
        case types.Integer(bitwidth=bw) | types.Float(bitwidth=bw):
            return bw
        case types.UniTuple(dtype=dtype):
            return _type_alignment_bits(dtype)
        case types.BaseTuple():
            layout = _llvm_struct_layout_bits(numba_type.types)
            return None if layout is None else layout[2]
        case _:
            return None


def _llvm_struct_layout_bits(field_types):
    """Compute LLVM literal-struct layout, including alignment padding."""
    member_offsets = []
    offset_bits = 0
    max_alignment_bits = _BYTE_SIZE_BITS
    for t in field_types:
        field_size_bits = _type_size_bits(t)
        field_alignment_bits = _type_alignment_bits(t)
        if field_size_bits is None or field_alignment_bits is None:
            return None
        offset_bits = _align_to_bits(offset_bits, field_alignment_bits)
        member_offsets.append(offset_bits)
        offset_bits += field_size_bits
        max_alignment_bits = max(max_alignment_bits, field_alignment_bits)
    total_size_bits = _align_to_bits(offset_bits, max_alignment_bits)
    return (member_offsets, total_size_bits, max_alignment_bits)


def _type_size_bits(numba_type):
    numba_type = _strip_literal_type(numba_type)
    match numba_type:
        case types.Boolean():
            return _BYTE_SIZE_BITS
        case types.Integer(bitwidth=bw) | types.Float(bitwidth=bw):
            return bw
        case types.UniTuple(dtype=dtype, count=count):
            elem_bits = _type_size_bits(dtype)
            return None if elem_bits is None else elem_bits * count
        case types.BaseTuple():
            layout = _llvm_struct_layout_bits(numba_type.types)
            return None if layout is None else layout[1]
        case types.Record():
            return numba_type.size * _BYTE_SIZE_BITS
        case _:
            return None


def _strip_literal_type(numba_type):
    return getattr(numba_type, "literal_type", numba_type)


def _uni_tuple_di_type(numba_type):
    elem_di = _numba_type_to_di_type_str(numba_type.dtype)
    elem_bits = _type_size_bits(numba_type.dtype)
    elem_str = _llvm_type_str(numba_type.dtype)
    if elem_di is None or elem_bits is None or elem_str is None:
        return None

    return _composite_di_type(
        "DW_TAG_array_type",
        name=f"UniTuple({numba_type.dtype} x {numba_type.count}) ([{numba_type.count} x {elem_str}])",
        base_type=elem_di,
        size_bits=numba_type.count * elem_bits,
        elements=[_subrange_di_type(numba_type.count)],
    )


def _base_tuple_di_type(numba_type):
    members_di = []
    llvm_member_types = []
    tuple_layout = _llvm_struct_layout_bits(numba_type.types)
    if tuple_layout is None:
        return None
    member_offsets, total_size_bits, _ = tuple_layout
    for i, (field_type, offset_bits) in enumerate(
        zip(numba_type.types, member_offsets, strict=True)
    ):
        field_di = _numba_type_to_di_type_str(field_type)
        field_bits = _type_size_bits(field_type)
        if field_di is None or field_bits is None:
            return None
        members_di.append(
            _derived_di_type(
                "DW_TAG_member",
                name=f"f{i}",
                base_type=field_di,
                offset_bits=offset_bits,
                size_bits=field_bits,
            )
        )
        llvm_str = _llvm_type_str(field_type)
        if llvm_str is None:
            return None
        llvm_member_types.append(llvm_str)
    type_name = ", ".join(str(t) for t in numba_type.types)
    llvm_type = ", ".join(llvm_member_types)
    return _composite_di_type(
        "DW_TAG_structure_type",
        name=f"Tuple({type_name}) ({{{llvm_type}}})",
        size_bits=total_size_bits,
        elements=members_di,
    )


def _record_di_type(numba_type):
    members_di = []
    for field_name in numba_type.fields:
        field_type = numba_type.typeof(field_name)
        field_di = _numba_type_to_di_type_str(field_type)
        field_bits = _type_size_bits(field_type)
        if field_di is None or field_bits is None:
            return None
        field_offset_bits = numba_type.offset(field_name) * _BYTE_SIZE_BITS
        members_di.append(
            _derived_di_type(
                "DW_TAG_member",
                name=field_name,
                base_type=field_di,
                offset_bits=field_offset_bits,
                size_bits=field_bits,
            )
        )
    return _composite_di_type(
        "DW_TAG_structure_type",
        name=str(numba_type),
        size_bits=numba_type.size * _BYTE_SIZE_BITS,
        elements=members_di,
    )


def _array_di_type(numba_type):
    elem = numba_type.dtype
    ndim = numba_type.ndim
    layout = numba_type.layout

    i8_di = _basic_di_type("i8", _BYTE_SIZE_BITS, "unsigned")
    i8_ptr_di = _derived_di_type(
        "DW_TAG_pointer_type",
        base_type=i8_di,
        size_bits=_POINTER_BITS,
    )

    member = lambda n, b, o, s: _derived_di_type(
        "DW_TAG_member",
        name=n,
        base_type=b,
        size_bits=s,
        offset_bits=o,
    )

    def field_debug_info(field_type):
        if isinstance(field_type, (types.MemInfoPointer, types.PyObject)):
            return i8_ptr_di, _POINTER_BITS, "i8*"
        if isinstance(field_type, types.CPointer):
            field_di = _cpointer_di_type(field_type)
            pointee_llvm = _llvm_type_str(field_type.dtype) or "i8"
            return field_di, _POINTER_BITS, f"{pointee_llvm}*"
        field_di = _numba_type_to_di_type_str(field_type)
        field_size = _type_size_bits(field_type)
        field_llvm = _llvm_type_str(field_type)
        return field_di, field_size, field_llvm

    members = []
    total_size = 0
    llvm_members = []
    for field_name, field_type in ArrayModel.get_members(numba_type):
        field_di, field_size, field_llvm = field_debug_info(field_type)
        if field_di is None or field_size is None or field_llvm is None:
            return None
        members.append(member(field_name, field_di, total_size, field_size))
        total_size += field_size
        llvm_members.append(field_llvm)
    llvm_struct = ", ".join(llvm_members)

    return _composite_di_type(
        "DW_TAG_structure_type",
        name=f"array({elem}, {ndim}d, {layout}) ({{{llvm_struct}}})",
        size_bits=total_size,
        elements=members,
    )


def _union_di_type(numba_type):
    variants = numba_type.types
    variant_sizes = [_type_size_bits(t) for t in variants]
    if not variant_sizes or any(bits is None for bits in variant_sizes):
        return None
    size_bits = max(variant_sizes)

    variant_members = []
    for tag, (variant_type, variant_bits) in enumerate(zip(variants, variant_sizes, strict=True)):
        variant_di = _numba_type_to_di_type_str(variant_type)
        if variant_di is None or variant_bits is None:
            return None
        variant_members.append(
            _derived_di_type(
                "DW_TAG_member",
                name=f"_{variant_type}",
                base_type=variant_di,
                size_bits=variant_bits,
                offset_bits=variant_bits,
                extra_data=tag,
            )
        )

    variant_key = ".".join(str(t) for t in variants)
    discriminator = _derived_di_type(
        "DW_TAG_member",
        name=f"discriminator-{variant_key}",
        base_type=_basic_di_type("uint8", _BYTE_SIZE_BITS, "unsigned"),
        size_bits=_BYTE_SIZE_BITS,
        flags="Artificial",
    )
    variant_part = _composite_di_type(
        "DW_TAG_variant_part",
        name="variant_part",
        size_bits=size_bits,
        elements=variant_members,
        identifier=f"variant-{variant_key}",
        discriminator=discriminator,
    )
    return _composite_di_type(
        "DW_TAG_structure_type",
        name="variant_wrapper_struct",
        size_bits=2 * size_bits,
        elements=[discriminator, variant_part],
        identifier=f"variant-wrapper-{variant_key}",
    )


def _numba_type_to_di_type_str(numba_type):
    """Map a Numba type to an MLIR #llvm.di_basic_type string."""
    numba_type = _strip_literal_type(numba_type)
    match numba_type:
        case types.Boolean():
            return _basic_di_type("bool", _BYTE_SIZE_BITS, "boolean")
        case types.Float(bitwidth=bw):
            return _basic_di_type(f"float{bw}", bw, "float")
        case types.Integer(signed=True, bitwidth=bw):
            return _basic_di_type(f"int{bw}", bw, "signed")
        case types.Integer(signed=False, bitwidth=bw):
            return _basic_di_type(f"uint{bw}", bw, "unsigned")
        case types.CPointer():
            return _cpointer_di_type(numba_type)
        case types.EnumMember():
            return _enum_member_di_type(numba_type)
        case types.NPDatetime():
            # NumPy stores datetime64 as signed int64.
            return _basic_di_type(f"datetime64[{numba_type.unit}]", _INT_LITERAL_BITS, "signed")
        case types.NPTimedelta():
            # NumPy stores timedelta64 as signed int64.
            return _basic_di_type(f"timedelta64[{numba_type.unit}]", _INT_LITERAL_BITS, "signed")
        case types.Complex(underlying_float=types.Float(bitwidth=bw)):
            return _complex_di_type(f"complex{bw * 2}", bw)
        case types.UniTuple():
            return _uni_tuple_di_type(numba_type)
        case types.BaseTuple():
            return _base_tuple_di_type(numba_type)
        case types.Record():
            return _record_di_type(numba_type)
        case types.Array():
            return _array_di_type(numba_type)
        case types.UnionType():
            return _union_di_type(numba_type)
        case Bfloat16():
            return _basic_di_type("__nv_bfloat16", _BFLOAT16_BITS, "float")
        case GridGroup():
            # GridGroup is an opaque cooperative-groups handle.
            return _basic_di_type("GridGroup", _GRID_GROUP_BITS, "unsigned")
        case _:
            return None


class DIBuilder:
    """Builds LLVM debug info MLIR attributes.

    All DI attributes are collected as text fragments and parsed in a single
    Module.parse call so that ``distinct[N]<>`` IDs resolve to the same
    DistinctAttr objects.
    """

    def __init__(self, loc, func_name, *, line_only=False, opt=True, context=None):
        self.context = context or numba_cuda_mlir_context.get_context()
        self._local_vars: list[_DILocalVarInfo] = []
        self.di_subprogram = None
        self.di_local_vars: dict[str, ir.Attribute] = {}
        self.di_expression = None
        self.arg_names: set[str] = set()

        raw_path = getattr(loc, "filename", None)
        if not raw_path or not raw_path.strip():
            self.valid = False
            return
        if raw_path.startswith("<") and raw_path.endswith(">"):
            # Synthetic paths like "<ipython-input-N>" (Jupyter cells) have no
            # file on disk, but IPython registers their source in Python's
            # linecache so DWARF can still reference meaningful line numbers.
            if not linecache.getlines(raw_path):
                self.valid = False
                return

        self.valid = True
        filename = os.path.basename(raw_path)
        directory = os.path.dirname(raw_path) or "."

        # Caller guarantees at least one of debug or lineinfo is True.
        emission = "DebugDirectivesOnly" if line_only else "Full"

        self._di_file_str = (
            f"#llvm.di_file<{_mlir_string_attr(filename)} in {_mlir_string_attr(directory)}>"
        )
        self._di_cu_str = (
            f"#llvm.di_compile_unit<"
            f"id = distinct[0]<>, "
            f"sourceLanguage = DW_LANG_C_plus_plus, "
            f"file = {self._di_file_str}, "
            f"producer = {_mlir_string_attr('clang (numba-cuda-mlir)')}, "
            f"isOptimized = {str(opt).lower()}, "
            f"emissionKind = {emission}"
            f">"
        )
        subprogram_flags = "Definition|Optimized" if opt else "Definition"
        self._di_subroutine_type_str = "#llvm.di_subroutine_type<types = #llvm.di_null_type>"
        self._di_sp_str = (
            f"#llvm.di_subprogram<"
            f"id = distinct[1]<>, "
            f"compileUnit = {self._di_cu_str}, "
            f"scope = {self._di_file_str}, "
            f"name = {_mlir_string_attr(func_name)}, "
            f"file = {self._di_file_str}, "
            f"type = {self._di_subroutine_type_str}, "
            f"line = {loc.line}, "
            f"scopeLine = {loc.line}, "
            f"subprogramFlags = {_mlir_string_attr(subprogram_flags)}"
            f">"
        )

    def add_local_variable(self, name, line, numba_type, *, arg_index=None):
        """Register a local variable for debug info emission."""
        if not self.valid:
            return
        type_str = _numba_type_to_di_type_str(numba_type)
        if type_str is None:
            return
        if arg_index is not None:
            self.arg_names.add(name)
        self._local_vars.append(_DILocalVarInfo(name, line, type_str, arg_index))

    def build(self):
        """Parse all DI metadata in a single call for consistent distinct IDs.

        Returns the di_subprogram attribute, or None if valid is False.
        """
        if not self.valid:
            return None

        attrs = {"numba_cuda_mlir.di_sp": self._di_sp_str}
        var_keys = []
        for i, var in enumerate(self._local_vars):
            arg_field = f", arg = {var.arg_index}" if var.arg_index is not None else ""
            var_str = (
                f"#llvm.di_local_variable<"
                f"scope = {self._di_sp_str}, "
                f"name = {_mlir_string_attr(var.name)}{arg_field}, "
                f"file = {self._di_file_str}, "
                f"line = {var.line}, "
                f"type = {var.type_str}"
                f">"
            )
            key = f"numba_cuda_mlir.di_var_{var.name}_{i}"
            attrs[key] = var_str
            var_keys.append((var.name, key))
        attrs["numba_cuda_mlir.di_expr"] = "#llvm.di_expression<>"

        attr_strs = ", ".join(f"{k} = {v}" for k, v in attrs.items())
        module_str = f"module attributes {{{attr_strs}}} {{}}"

        with self.context:
            helper = ir.Module.parse(module_str)
            self.di_subprogram = helper.operation.attributes["numba_cuda_mlir.di_sp"]
            self.di_expression = helper.operation.attributes["numba_cuda_mlir.di_expr"]
            for name, key in var_keys:
                self.di_local_vars[name] = helper.operation.attributes[key]

        return self.di_subprogram



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Optional

from ...ir import Attribute, Operation, Value, StringAttr
from .._transform_debug_extension_ops_gen import *
from .._transform_pdl_extension_ops_gen import _Dialect

try:
    from .._ods_common import _cext as _ods_cext
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from typing import Union


@_ods_cext.register_operation(_Dialect, replace=True)
class EmitParamAsRemarkOp(EmitParamAsRemarkOp):
    def __init__(
        self,
        param: Attribute,
        *,
        anchor: Optional[Operation] = None,
        message: Optional[Union[StringAttr, str]] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(message, str):
            message = StringAttr.get(message)

        super().__init__(
            param,
            anchor=anchor,
            message=message,
            loc=loc,
            ip=ip,
        )


def emit_param_as_remark(
    param: Attribute,
    *,
    anchor: Optional[Operation] = None,
    message: Optional[Union[StringAttr, str]] = None,
    loc=None,
    ip=None,
):
    return EmitParamAsRemarkOp(param, anchor=anchor, message=message, loc=loc, ip=ip)


@_ods_cext.register_operation(_Dialect, replace=True)
class EmitRemarkAtOp(EmitRemarkAtOp):
    def __init__(
        self,
        at: Union[Operation, Value],
        message: Optional[Union[StringAttr, str]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if isinstance(message, str):
            message = StringAttr.get(message)

        super().__init__(
            at,
            message,
            loc=loc,
            ip=ip,
        )


def emit_remark_at(
    at: Union[Operation, Value],
    message: Optional[Union[StringAttr, str]] = None,
    *,
    loc=None,
    ip=None,
):
    return EmitRemarkAtOp(at, message, loc=loc, ip=ip)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Autogenerated by numba_cuda_mlir/runtime/stubgen.py
# Please do not edit

from numba_cuda_mlir.typing import (
    none,
    ptr,
    int8,
    int16,
    int32,
    int64,
    bool,
    float32,
    float64,
    UniTuple,
    Tuple,
)

def elect_sync() -> bool: ...
def elect_sync_mask(arg0: int32) -> bool: ...
def mbarrier_init(arg0: ptr, arg1: int32) -> None: ...
def mbarrier_try_wait(arg0: ptr, arg1: int32) -> None: ...
def barrier() -> None: ...
def mbarrier_arrive_expect_tx(arg0: ptr, arg1: int32) -> None: ...
def cp_async_bulk_load(arg0: ptr, arg1: ptr, arg2: ptr, arg3: int32) -> None: ...
def fence_mbarrier_init() -> None: ...
def cp_async_bulk_store(arg0: ptr, arg1: ptr, arg2: int32) -> None: ...
def cp_async_bulk_tensor_2d(arg0: ptr, arg1: ptr, arg2: int32, arg3: int32, arg4: ptr) -> None: ...
def cp_async_bulk_tensor_2d_multicast(
    arg0: ptr, arg1: ptr, arg2: int32, arg3: int32, arg4: ptr, arg5: int16
) -> None: ...
def cluster_arrive_relaxed() -> None: ...
def cluster_wait() -> None: ...
def block_idx_in_cluster() -> int32: ...
def cluster_ctarank() -> int32: ...
def mapa(arg0: ptr, arg1: int32) -> ptr: ...
def warp_idx() -> int32: ...
def prefetch_tensormap(arg0: ptr) -> None: ...
def encode_instr_descriptor(
    arg0: int16,
    arg1: bool,
    arg2: bool,
    arg3: int16,
    arg4: bool,
    arg5: int16,
    arg6: int16,
    arg7: bool,
    arg8: bool,
    arg9: bool,
    arg10: bool,
    arg11: int16,
    arg12: int16,
    arg13: int16,
) -> int32: ...
def encode_smem_descriptor(
    arg0: int16, arg1: int16, arg2: int16, arg3: int8, arg4: int8
) -> int64: ...
def breakpoint() -> None: ...
def nanosleep(arg0: int32) -> None: ...
def round_ndigits_type_f32_type_i32(arg0: float32, arg1: int32) -> float32: ...
def round_ndigits_type_f32_type_i64(arg0: float32, arg1: int64) -> float32: ...
def round_ndigits_type_f64_type_i32(arg0: float64, arg1: int32) -> float64: ...
def round_ndigits_type_f64_type_i64(arg0: float64, arg1: int64) -> float64: ...
def __nv_ffs(arg0: int32) -> int32: ...
def __nv_ffsll(arg0: int64) -> int32: ...
def ffs(arg0: int32) -> int32: ...
def ffsll(arg0: int64) -> int32: ...
def tcgen05_alloc(arg0: ptr, arg1: int32) -> None: ...
def tcgen05_mma(arg0: int32, arg1: int64, arg2: int64, arg3: int32, arg4: int32) -> None: ...
def tcgen05_commit_arrive(arg0: ptr) -> None: ...
def tcgen05_ld(arg0: int32) -> float32: ...
def tcgen05_ld_x2(arg0: int32) -> tuple[float32, float32]: ...
def tcgen05_dealloc(arg0: int32, arg1: int32) -> None: ...
def tcgen05_relinquish_alloc_permit() -> None: ...
def tcgen05_alloc_2cta(arg0: ptr, arg1: int32) -> None: ...
def tcgen05_mma_2cta(arg0: int32, arg1: int64, arg2: int64, arg3: int32, arg4: int32) -> None: ...
def tcgen05_dealloc_2cta(arg0: int32, arg1: int32) -> None: ...
def tcgen05_relinquish_alloc_permit_2cta() -> None: ...



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from typing import Dict, List, Sequence, Union

from contextlib import contextmanager
import functools
import inspect
import threading

from ..... import ir
from ...._ods_common import (
    get_op_result_or_value as _get_op_result_or_value,
    get_op_results_or_values as _get_op_results_or_values,
)
from .comprehension import *
from .config import *
from .emitter import *

_CONTEXT = threading.local()

StructuredOpOuts = Union[
    ir.Operation,
    ir.OpView,
    ir.OpResultList,
    Sequence[Union[ir.Value, ir.Operation, ir.OpView]],
]


@contextmanager
def bind_op_def(op_def: LinalgOpDef):
    if hasattr(_CONTEXT, "current_op_def"):
        raise ValueError("Cannot recursively define an operation")
    _CONTEXT.current_op_def = op_def
    try:
        yield op_def
    finally:
        del _CONTEXT.current_op_def


def current_op_def() -> LinalgOpDef:
    try:
        return _CONTEXT.current_op_def
    except AttributeError:
        raise ValueError(
            "Attempt to access the current op definition being defined "
            "but none is set. Did you mean to call this in an op definition?"
        )


def _prepare_structured_op_outs(outs: StructuredOpOuts) -> ValueList:
    if isinstance(outs, (ir.Operation, ir.OpView)):
        return _get_op_results_or_values(outs)
    elif isinstance(outs, ir.OpResultList):
        return outs

    return [_get_op_result_or_value(o) for o in outs]


class DefinedOpCallable:
    """Callable that wraps any defined op function."""

    def __init__(self, op_name: str, op_def: LinalgOpDef):
        self.op_name = op_name
        self.op_def = op_def

    def __call__(
        self,
        *ins: Union[ir.Operation, ir.OpView, ir.Value],
        outs: StructuredOpOuts,
        **kwargs,
    ):
        """Emits the corresponding op definition as IR.

        Most arguments are passed through to the underlying emitter. The following
        keyword argument is interpreted here:
          emit_generic: Emits a generic form as appropriate (default True). If
            False, a named form is emitted (which must have been built in to the
            compiler).
        """
        emit_generic = kwargs.pop("emit_generic", False)
        if not isinstance(emit_generic, bool):
            raise ValueError(
                f"The named argument 'emit_generic' needs to be "
                f" of type bool but got {type(emit_generic)}"
            )

        op_configs = LinalgOpConfig.from_linalg_op_def(
            self.op_def, context=ir.Context.current
        )

        if len(op_configs) != 1:
            # TODO: Support composite ops.
            raise NotImplementedError(
                f"Emission of composite linalg ops not supported: {op_configs}"
            )

        ctx = ir.Context.current
        linalgDialect = ctx.get_dialect_descriptor("linalg")
        fully_qualified_name = "linalg." + self.op_name
        emit_generic = emit_generic or not ctx.is_registered_operation(
            fully_qualified_name
        )

        op_config = op_configs[0]
        out_values = _prepare_structured_op_outs(outs)
        in_values = [_get_op_result_or_value(i) for i in ins]
        if op_config.structured_op:
            if emit_generic:
                return emit_generic_structured_op(
                    op_config.structured_op, *in_values, outs=out_values, **kwargs
                )
            else:
                return emit_named_structured_op(
                    op_config.structured_op,
                    self.op_name,
                    self.op_def.metadata.cpp_class_name,
                    *in_values,
                    outs=out_values,
                    **kwargs,
                )

        raise NotImplementedError(
            f"Emission of linalg op type not supported: {op_config}"
        )


def linalg_structured_op(
    dsl_func=None, *, op_name=None, op_class_name=None
) -> DefinedOpCallable:
    if dsl_func is None:
        # Curry the keyword args in for delayed application.
        return functools.partial(
            linalg_structured_op, op_name=op_name, op_class_name=op_class_name
        )
    # Determine default names by introspecting the function.
    if op_name is None:
        op_name = dsl_func.__name__
    if op_class_name is None:
        # Camel case it.
        op_class_name = f"{''.join(x.title() for x in op_name.split('_'))}Op"

    op_def = LinalgOpDef(
        name=op_name, cpp_class_name=op_class_name, doc=inspect.getdoc(dsl_func)
    )

    # Extract arguments and TensorDefs from the signature.
    dsl_func_args = list()
    sig = inspect.signature(dsl_func)
    for param_name, param in sig.parameters.items():
        param_default = param.default
        if isinstance(
            param_default,
            (
                TensorDef,
                ScalarDef,
                IndexAttrDef,
                UnaryFnAttrDef,
                BinaryFnAttrDef,
                TypeFnAttrDef,
            ),
        ):
            op_def.add_operand(param_name, param_default.operand_def)
        else:
            raise ValueError(
                f"@linalg_structured_op function parameters must be defaulted as "
                f"TensorDef(...), ScalarDef(...), or IndexAttrDef(...): "
                f"Found {param_name}: {param_default}"
            )
        dsl_func_args.append(param_default)

    # Invoke the DSL func to finish populating the op definition.
    with bind_op_def(op_def):
        dsl_func(*dsl_func_args)

    # TODO: The returned callable should be an IR emitter but that is not
    # upstreamed yet.
    return DefinedOpCallable(op_name, op_def)


def domain(*dimensions: DimDef):
    if any(not isinstance(d, DimDef) for d in dimensions):
        raise ValueError(f"Expected dimensions of type DimDef but got {dimensions}")
    current_op_def().domain.extend(dimensions)


def implements(*interfaces: OpInterfaceDef):
    if any(not isinstance(intr, OpInterfaceDef) for intr in interfaces):
        raise ValueError(
            f"Expected interfaces of type OpInterfaceDef but got {interfaces}"
        )
    current_op_def().metadata.implements.extend(interfaces)


def defines(*definitions: OpDefinitionDef):
    if any(not isinstance(defi, OpDefinitionDef) for defi in definitions):
        raise ValueError(
            f"Expected definitions of type OpDefinitionDef but got {definitions}"
        )
    current_op_def().metadata.defines.extend(definitions)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.types import *  # noqa: F401,F403
from numba_cuda_mlir.types import *  # noqa: F401,F403



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
from numba_cuda_mlir.mlir_lowering import MLIRLower
from numba_cuda_mlir.lowering_utilities import index_of, memref_to_llvm_ptr, convert, get_type_width
from numba_cuda_mlir.lowering_utilities.type_conversions import to_mlir_type
from numba_cuda_mlir.cuda import vector as vector_module
from numba_cuda_mlir import types
from numba_cuda_mlir.type_defs.vector_types import VectorType
from numba_cuda_mlir._mlir.dialects import vector, arith, llvm
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir import ir
from typing import Any

# For llvm.getelementptr - dynamic offset marker
llvm_kDynamic = -2147483648


def _cpointer_to_element_ptr(ptr: ir.Value, index: ir.Value, element_type: ir.Type) -> ir.Value:
    """Convert CPointer + index to element pointer using getelementptr."""
    idx = convert(index, T.i64())
    return llvm.getelementptr(
        llvm.PointerType.get(),
        ptr,
        [idx],
        [llvm_kDynamic],
        element_type,
        None,
    )


def _get_alignment(lower: MLIRLower, args: list[Any], kwargs) -> int:
    """Extract alignment value from positional args or kwargs."""
    if len(args) > 3:
        # Passed as positional argument (4th arg)
        return lower.get_numba_type(args[3].name).literal_value
    else:
        # Search in keyword arguments
        kwargs_dict = dict(kwargs)
        if "alignment" not in kwargs_dict:
            raise ValueError("alignment parameter not found in args or kwargs")
        return lower.get_numba_type(kwargs_dict["alignment"].name).literal_value


def _get_padding_value(element_type: ir.Type) -> ir.Value:
    """Get a zero padding value for the given element type."""
    if isinstance(element_type, ir.FloatType):
        return arith.constant(element_type, 0.0)
    elif isinstance(element_type, ir.IntegerType):
        return arith.constant(element_type, 0)
    else:
        raise NotImplementedError(f"Unsupported element type for vector: {element_type}")


def _get_permutation_map(memref_rank: int, vec_rank: int) -> ir.AffineMap:
    """Get a permutation map for vector transfer ops.

    For a memref of rank M and vector of rank V, the permutation map
    maps from M input dimensions to V output dimensions.
    We map the last V dimensions of the memref to the vector.

    e.g., for memref<8x4xf32> and vector<4xf32>:
    permutation_map = (d0, d1) -> (d1)
    """
    offset = memref_rank - vec_rank
    dims = [ir.AffineDimExpr.get(offset + i) for i in range(vec_rank)]
    return ir.AffineMap.get(memref_rank, 0, dims)


def _vector_with_element_type(vec_type: ir.VectorType, element_type: ir.Type) -> ir.VectorType:
    return ir.VectorType.get(list(ir.VectorType(vec_type).shape), element_type)


def _integer_bits_vector_type(value_vec_type: ir.VectorType) -> ir.VectorType:
    value_vec_type = ir.VectorType(value_vec_type)
    return _vector_with_element_type(
        value_vec_type,
        ir.IntegerType.get_signless(get_type_width(value_vec_type.element_type)),
    )


def _array_storage_vector_type(array: ir.Value, value_vec_type: ir.VectorType) -> ir.VectorType:
    return _vector_with_element_type(value_vec_type, array.type.element_type)


def _scalar_storage_vector_type(lower: MLIRLower, vec_numba_type: VectorType) -> ir.VectorType:
    value_vec_type = lower.get_mlir_type(vec_numba_type)
    return _vector_with_element_type(value_vec_type, lower.get_storage_type(vec_numba_type.dtype))


def _storage_vector_to_value(
    vec_numba_type: VectorType, value: ir.Value, value_vec_type: ir.VectorType
):
    if value.type == value_vec_type:
        return value

    storage_vec_type = ir.VectorType(value.type)
    value_vec_type = ir.VectorType(value_vec_type)
    storage_elem_type = storage_vec_type.element_type
    value_elem_type = value_vec_type.element_type

    if isinstance(vec_numba_type.dtype, (types.Boolean, types.BooleanLiteral)):
        from numba_cuda_mlir.mlir.dialect_exts import arith as arith_ext

        zero = arith_ext.constant(0, storage_vec_type)
        return arith.cmpi(arith.CmpIPredicate.ne, value, zero)

    if isinstance(value_elem_type, ir.FloatType) and isinstance(storage_elem_type, ir.IntegerType):
        bits_type = _integer_bits_vector_type(value_vec_type)
        bits = value
        bits_elem_type = ir.VectorType(bits_type).element_type
        if storage_elem_type.width > bits_elem_type.width:
            bits = arith.trunci(out=bits_type, in_=value)
        elif storage_elem_type.width < bits_elem_type.width:
            bits = arith.extui(out=bits_type, in_=value)
        elif value.type != bits_type:
            bits = arith.bitcast(out=bits_type, in_=value)
        return arith.bitcast(out=value_vec_type, in_=bits)

    return convert(value, value_vec_type)


def _value_vector_to_storage(
    vec_numba_type: VectorType, value: ir.Value, storage_vec_type: ir.VectorType
):
    if value.type == storage_vec_type:
        return value

    value_vec_type = ir.VectorType(value.type)
    storage_vec_type = ir.VectorType(storage_vec_type)
    value_elem_type = value_vec_type.element_type
    storage_elem_type = storage_vec_type.element_type

    if isinstance(vec_numba_type.dtype, (types.Boolean, types.BooleanLiteral)):
        return arith.extui(out=storage_vec_type, in_=value)

    if isinstance(value_elem_type, ir.FloatType) and isinstance(storage_elem_type, ir.IntegerType):
        bits_type = _integer_bits_vector_type(value_vec_type)
        bits = arith.bitcast(out=bits_type, in_=value)
        bits_elem_type = ir.VectorType(bits_type).element_type
        if bits_elem_type.width == storage_elem_type.width:
            return (
                bits
                if bits_type == storage_vec_type
                else arith.bitcast(out=storage_vec_type, in_=bits)
            )
        if bits_elem_type.width < storage_elem_type.width:
            return arith.extui(out=storage_vec_type, in_=bits)
        return arith.trunci(out=storage_vec_type, in_=bits)

    return convert(value, storage_vec_type)


# =============================================================================
# Unaligned Vector Load (3-arg) - uses vector.transfer_read
# =============================================================================


@lower(vector_module.load, types.Array, types.Integer, types.IntegerLiteral)
@lower(vector_module.load, types.Array, types.Integer, types.Tuple)
@lower(vector_module.load, types.Array, types.Integer, types.UniTuple)
def vector_load_1d_index(lower: MLIRLower, target, args: list[Any], kwargs):
    """Unaligned vector load."""
    array = lower.load_var(args[0])
    index = lower.load_var(args[1])

    target_type = lower.get_numba_type(target.name)
    value_vec_type = lower.get_mlir_type(target_type)
    vec_type = _array_storage_vector_type(array, value_vec_type)
    vec_rank = len(value_vec_type.shape)
    memref_rank = array.type.rank

    index = index_of(index)
    padding = _get_padding_value(vec_type.element_type)
    perm_map = _get_permutation_map(memref_rank, vec_rank)
    in_bounds = ir.ArrayAttr.get([ir.BoolAttr.get(True)] * vec_rank)
    result = vector.transfer_read(vec_type, array, [index], perm_map, padding, in_bounds)
    lower.store_var(target, _storage_vector_to_value(target_type, result, value_vec_type))


@lower(vector_module.load, types.Array, types.UniTuple, types.IntegerLiteral)
@lower(vector_module.load, types.Array, types.UniTuple, types.Tuple)
@lower(vector_module.load, types.Array, types.UniTuple, types.UniTuple)
@lower(vector_module.load, types.Array, types.Tuple, types.IntegerLiteral)
@lower(vector_module.load, types.Array, types.Tuple, types.Tuple)
@lower(vector_module.load, types.Array, types.Tuple, types.UniTuple)
def vector_load_nd_index(lower: MLIRLower, target, args: list[Any], kwargs):
    """Unaligned vector load."""
    array = lower.load_var(args[0])
    indices = lower.load_var(args[1])

    target_type = lower.get_numba_type(target.name)
    value_vec_type = lower.get_mlir_type(target_type)
    vec_type = _array_storage_vector_type(array, value_vec_type)
    vec_rank = len(value_vec_type.shape)
    memref_rank = array.type.rank

    indices = [index_of(i) for i in indices]
    padding = _get_padding_value(vec_type.element_type)
    perm_map = _get_permutation_map(memref_rank, vec_rank)
    in_bounds = ir.ArrayAttr.get([ir.BoolAttr.get(True)] * vec_rank)
    result = vector.transfer_read(vec_type, array, indices, perm_map, padding, in_bounds)
    lower.store_var(target, _storage_vector_to_value(target_type, result, value_vec_type))


# =============================================================================
# Aligned Vector Load (4-arg) - uses llvm.load with alignment
# =============================================================================


@lower(
    vector_module.load,
    types.Array,
    types.Integer,
    types.IntegerLiteral,
    types.IntegerLiteral,
)
@lower(vector_module.load, types.Array, types.Integer, types.Tuple, types.IntegerLiteral)
@lower(vector_module.load, types.Array, types.Integer, types.UniTuple, types.IntegerLiteral)
def vector_load_1d_index_aligned(lower: MLIRLower, target, args: list[Any], kwargs):
    """Aligned vector load using llvm.load (1D index)."""
    array = lower.load_var(args[0])
    index = lower.load_var(args[1])

    target_type = lower.get_numba_type(target.name)
    value_vec_type = lower.get_mlir_type(target_type)
    vec_type = _array_storage_vector_type(array, value_vec_type)
    index = index_of(index)

    # Convert memref to LLVM pointer
    ptr = memref_to_llvm_ptr(array, [index], vec_type.element_type)

    # Use llvm.load with alignment
    alignment = _get_alignment(lower, args, kwargs)
    result = llvm.load(vec_type, ptr, alignment=alignment)
    lower.store_var(target, _storage_vector_to_value(target_type, result, value_vec_type))


@lower(
    vector_module.load,
    types.Array,
    types.UniTuple,
    types.IntegerLiteral,
    types.IntegerLiteral,
)
@lower(vector_module.load, types.Array, types.UniTuple, types.Tuple, types.IntegerLiteral)
@lower(
    vector_module.load,
    types.Array,
    types.UniTuple,
    types.UniTuple,
    types.IntegerLiteral,
)
@lower(
    vector_module.load,
    types.Array,
    types.Tuple,
    types.IntegerLiteral,
    types.IntegerLiteral,
)
@lower(vector_module.load, types.Array, types.Tuple, types.Tuple, types.IntegerLiteral)
@lower(vector_module.load, types.Array, types.Tuple, types.UniTuple, types.IntegerLiteral)
def vector_load_nd_index_aligned(lower: MLIRLower, target, args: list[Any], kwargs):
    """Aligned vector load using llvm.load (N-D index)."""
    array = lower.load_var(args[0])
    indices = lower.load_var(args[1])

    target_type = lower.get_numba_type(target.name)
    value_vec_type = lower.get_mlir_type(target_type)
    vec_type = _array_storage_vector_type(array, value_vec_type)
    indices = [index_of(i) for i in indices]

    # Convert memref to LLVM pointer
    ptr = memref_to_llvm_ptr(array, indices, vec_type.element_type)

    # Use llvm.load with alignment
    alignment = _get_alignment(lower, args, kwargs)
    result = llvm.load(vec_type, ptr, alignment=alignment)
    lower.store_var(target, _storage_vector_to_value(target_type, result, value_vec_type))


# =============================================================================
# Unaligned Vector Store (3-arg) - uses vector.transfer_write
# =============================================================================


@lower(vector_module.store, types.Array, types.Integer, VectorType)
def vector_store_1d_index(lower: MLIRLower, target, args: list[Any], kwargs):
    """Unaligned vector store."""
    array = lower.load_var(args[0])
    index = lower.load_var(args[1])
    vec = lower.load_var(args[2])

    vec_numba_type = lower.get_numba_type(args[2].name)
    vec_type = _array_storage_vector_type(array, lower.get_mlir_type(vec_numba_type))
    vec = _value_vector_to_storage(vec_numba_type, vec, vec_type)
    vec_rank = len(vec_type.shape)
    memref_rank = array.type.rank

    index = index_of(index)
    perm_map = _get_permutation_map(memref_rank, vec_rank)
    in_bounds = ir.ArrayAttr.get([ir.BoolAttr.get(True)] * vec_rank)
    vector.transfer_write(None, vec, array, [index], perm_map, in_bounds)
    lower.store_var(target, None)


@lower(vector_module.store, types.Array, types.UniTuple, VectorType)
@lower(vector_module.store, types.Array, types.Tuple, VectorType)
def vector_store_nd_index(lower: MLIRLower, target, args: list[Any], kwargs):
    """Unaligned vector store."""
    array = lower.load_var(args[0])
    indices = lower.load_var(args[1])
    vec = lower.load_var(args[2])

    vec_numba_type = lower.get_numba_type(args[2].name)
    vec_type = _array_storage_vector_type(array, lower.get_mlir_type(vec_numba_type))
    vec = _value_vector_to_storage(vec_numba_type, vec, vec_type)
    vec_rank = len(vec_type.shape)
    memref_rank = array.type.rank

    indices = [index_of(i) for i in indices]
    perm_map = _get_permutation_map(memref_rank, vec_rank)
    in_bounds = ir.ArrayAttr.get([ir.BoolAttr.get(True)] * vec_rank)
    vector.transfer_write(None, vec, array, indices, perm_map, in_bounds)
    lower.store_var(target, None)


# =============================================================================
# Aligned Vector Store (4-arg) - uses llvm.store with alignment
# =============================================================================


@lower(vector_module.store, types.Array, types.Integer, VectorType, types.IntegerLiteral)
def vector_store_1d_index_aligned(lower: MLIRLower, target, args: list[Any], kwargs):
    """Aligned vector store using llvm.store (1D index)."""
    array = lower.load_var(args[0])
    index = lower.load_var(args[1])
    vec = lower.load_var(args[2])
    index = index_of(index)

    vec_numba_type = lower.get_numba_type(args[2].name)
    vec_type = _array_storage_vector_type(array, lower.get_mlir_type(vec_numba_type))
    vec = _value_vector_to_storage(vec_numba_type, vec, vec_type)

    # Convert memref to LLVM pointer
    ptr = memref_to_llvm_ptr(array, [index], vec_type.element_type)

    # Use llvm.store with alignment
    alignment = _get_alignment(lower, args, kwargs)
    llvm.store(vec, ptr, alignment=alignment)
    lower.store_var(target, None)


@lower(vector_module.store, types.Array, types.UniTuple, VectorType, types.IntegerLiteral)
@lower(vector_module.store, types.Array, types.Tuple, VectorType, types.IntegerLiteral)
def vector_store_nd_index_aligned(lower: MLIRLower, target, args: list[Any], kwargs):
    """Aligned vector store using llvm.store (N-D index)."""
    array = lower.load_var(args[0])
    indices = lower.load_var(args[1])
    vec = lower.load_var(args[2])
    indices = [index_of(i) for i in indices]

    vec_numba_type = lower.get_numba_type(args[2].name)
    vec_type = _array_storage_vector_type(array, lower.get_mlir_type(vec_numba_type))
    vec = _value_vector_to_storage(vec_numba_type, vec, vec_type)

    # Convert memref to LLVM pointer
    ptr = memref_to_llvm_ptr(array, indices, vec_type.element_type)

    # Use llvm.store with alignment
    alignment = _get_alignment(lower, args, kwargs)
    llvm.store(vec, ptr, alignment=alignment)
    lower.store_var(target, None)


# =============================================================================
# Aligned CPointer operations - uses llvm.load/store with alignment
# =============================================================================


@lower(
    vector_module.load,
    types.CPointer,
    types.Integer,
    types.IntegerLiteral,
    types.IntegerLiteral,
)
@lower(
    vector_module.load,
    types.CPointer,
    types.Integer,
    types.Tuple,
    types.IntegerLiteral,
)
@lower(
    vector_module.load,
    types.CPointer,
    types.Integer,
    types.UniTuple,
    types.IntegerLiteral,
)
def vector_load_cpointer_1d(lower: MLIRLower, target, args: list[Any], kwargs):
    """Vector load from CPointer using llvm.load with alignment (1D index)."""
    ptr = lower.load_var(args[0])
    index = lower.load_var(args[1])

    target_type = lower.get_numba_type(target.name)
    value_vec_type = lower.get_mlir_type(target_type)
    vec_type = _scalar_storage_vector_type(lower, target_type)

    element_ptr = _cpointer_to_element_ptr(ptr, index, vec_type.element_type)

    alignment = _get_alignment(lower, args, kwargs)
    result = llvm.load(vec_type, element_ptr, alignment=alignment)
    lower.store_var(target, _storage_vector_to_value(target_type, result, value_vec_type))


@lower(vector_module.store, types.CPointer, types.Integer, VectorType, types.IntegerLiteral)
def vector_store_cpointer_1d(lower: MLIRLower, target, args: list[Any], kwargs):
    """Vector store to CPointer using llvm.store with alignment (1D index)."""
    ptr = lower.load_var(args[0])
    index = lower.load_var(args[1])
    vec = lower.load_var(args[2])

    vec_numba_type = lower.get_numba_type(args[2].name)
    vec_type = _scalar_storage_vector_type(lower, vec_numba_type)
    vec = _value_vector_to_storage(vec_numba_type, vec, vec_type)
    element_ptr = _cpointer_to_element_ptr(ptr, index, vec_type.element_type)

    alignment = _get_alignment(lower, args, kwargs)
    llvm.store(vec, element_ptr, alignment=alignment)
    lower.store_var(target, None)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import namedtuple, OrderedDict
import dis
import inspect

from types import CodeType, ModuleType

from numba_cuda_mlir.numba_cuda.core import errors
from numba_cuda_mlir.numba_cuda import serialize
from numba_cuda_mlir.numba_cuda import utils
from numba_cuda_mlir.numba_cuda.utils import PYVERSION
from numba_cuda_mlir._threading import _LockedCounter


if PYVERSION in ((3, 12), (3, 13), (3, 14)):
    from opcode import _inline_cache_entries

    # Instruction/opcode length in bytes
    INSTR_LEN = 2
elif PYVERSION in ((3, 9), (3, 10), (3, 11)):
    pass
else:
    raise NotImplementedError(PYVERSION)


opcode_info = namedtuple("opcode_info", ["argsize"])
_ExceptionTableEntry = namedtuple("_ExceptionTableEntry", "start end target depth lasti")

# The following offset is used as a hack to inject a NOP at the start of the
# bytecode. So that function starting with `while True` will not have block-0
# as a jump target. The Lowerer puts argument initialization at block-0.
_FIXED_OFFSET = 2


def get_function_object(obj):
    """
    Objects that wraps function should provide a "__numba__" magic attribute
    that contains a name of an attribute that contains the actual python
    function object.
    """
    attr = getattr(obj, "__numba__", None)
    if attr:
        return getattr(obj, attr)
    return obj


def get_code_object(obj):
    "Shamelessly borrowed from llpython"
    return getattr(obj, "__code__", getattr(obj, "func_code", None))


def _as_opcodes(seq):
    lst = []
    for s in seq:
        c = dis.opmap.get(s)
        if c is not None:
            lst.append(c)
    return lst


JREL_OPS = frozenset(dis.hasjrel)
JABS_OPS = frozenset(dis.hasjabs)
JUMP_OPS = JREL_OPS | JABS_OPS
TERM_OPS = frozenset(_as_opcodes(["RETURN_VALUE", "RAISE_VARARGS"]))
EXTENDED_ARG = dis.EXTENDED_ARG
HAVE_ARGUMENT = dis.HAVE_ARGUMENT


class ByteCodeInst:
    """
    Attributes
    ----------
    - offset:
        byte offset of opcode
    - opcode:
        opcode integer value
    - arg:
        instruction arg
    - lineno:
        -1 means unknown
    """

    __slots__ = "offset", "next", "opcode", "opname", "arg", "lineno"

    def __init__(self, offset, opcode, arg, nextoffset):
        self.offset = offset
        self.next = nextoffset
        self.opcode = opcode
        self.opname = dis.opname[opcode]
        self.arg = arg
        self.lineno = -1  # unknown line number

    @property
    def is_jump(self):
        return self.opcode in JUMP_OPS

    @property
    def is_terminator(self):
        return self.opcode in TERM_OPS

    def get_jump_target(self):
        # With Python 3.10 the addressing of "bytecode" instructions has
        # changed from using bytes to using 16-bit words instead. As a
        # consequence the code to determine where a jump will lead had to be
        # adapted.
        # See also:
        # https://bugs.python.org/issue26647
        # https://bugs.python.org/issue27129
        # https://github.com/python/cpython/pull/25069
        assert self.is_jump
        if PYVERSION in ((3, 13), (3, 14)):
            if self.opcode in (
                dis.opmap[k] for k in ["JUMP_BACKWARD", "JUMP_BACKWARD_NO_INTERRUPT"]
            ):
                return self.next - (self.arg * 2)
        elif PYVERSION in ((3, 12),):
            if self.opcode in (dis.opmap[k] for k in ["JUMP_BACKWARD"]):
                return self.offset - (self.arg - 1) * 2
        elif PYVERSION in ((3, 11),):
            if self.opcode in (
                dis.opmap[k]
                for k in (
                    "JUMP_BACKWARD",
                    "POP_JUMP_BACKWARD_IF_TRUE",
                    "POP_JUMP_BACKWARD_IF_FALSE",
                    "POP_JUMP_BACKWARD_IF_NONE",
                    "POP_JUMP_BACKWARD_IF_NOT_NONE",
                )
            ):
                return self.offset - (self.arg - 1) * 2
        elif PYVERSION in (
            (3, 9),
            (3, 10),
        ):
            pass
        else:
            raise NotImplementedError(PYVERSION)

        if PYVERSION in ((3, 10), (3, 11), (3, 12), (3, 13), (3, 14)):
            if self.opcode in JREL_OPS:
                return self.next + self.arg * 2
            else:
                assert self.opcode in JABS_OPS
                return self.arg * 2 - 2
        elif PYVERSION in ((3, 9),):
            if self.opcode in JREL_OPS:
                return self.next + self.arg
            else:
                assert self.opcode in JABS_OPS
                return self.arg
        else:
            raise NotImplementedError(PYVERSION)

    def __repr__(self):
        return "%s(arg=%s, lineno=%d)" % (self.opname, self.arg, self.lineno)

    @property
    def block_effect(self):
        """Effect of the block stack
        Returns +1 (push), 0 (none) or -1 (pop)
        """
        if self.opname.startswith("SETUP_"):
            return 1
        elif self.opname == "POP_BLOCK":
            return -1
        else:
            return 0


CODE_LEN = 1
ARG_LEN = 1
NO_ARG_LEN = 1

OPCODE_NOP = dis.opname.index("NOP")


if PYVERSION in ((3, 13), (3, 14)):

    def _unpack_opargs(code):
        buf = [(start_offset, op, arg) for _, start_offset, op, arg in dis._unpack_opargs(code)]
        for i, (start_offset, op, arg) in enumerate(buf):
            if i + 1 < len(buf):
                next_offset = buf[i + 1][0]
            else:
                next_offset = len(code)
            yield (start_offset, op, arg, next_offset)

elif PYVERSION in ((3, 9), (3, 10), (3, 11), (3, 12)):
    # Adapted from Lib/dis.py
    def _unpack_opargs(code):
        """
        Returns a 4-int-tuple of
        (bytecode offset, opcode, argument, offset of next bytecode).
        """
        extended_arg = 0
        n = len(code)
        offset = i = 0
        while i < n:
            op = code[i]
            i += CODE_LEN
            if op >= HAVE_ARGUMENT:
                arg = code[i] | extended_arg
                for j in range(ARG_LEN):
                    arg |= code[i + j] << (8 * j)
                i += ARG_LEN
                if PYVERSION in ((3, 12),):
                    # Python 3.12 introduced cache slots. We need to account for
                    # cache slots when we determine the offset of the next
                    # opcode. The number of cache slots is specific to each
                    # opcode and can be looked up in the _inline_cache_entries
                    # dictionary.
                    i += _inline_cache_entries[op] * INSTR_LEN
                elif PYVERSION in ((3, 9), (3, 10), (3, 11)):
                    pass
                else:
                    raise NotImplementedError(PYVERSION)
                if op == EXTENDED_ARG:
                    # This is a deviation from what dis does...
                    # In python 3.11 it seems like EXTENDED_ARGs appear more
                    # often and are also used as jump targets. So as to not have
                    # to do "book keeping" for where EXTENDED_ARGs have been
                    # "skipped" they are replaced with NOPs so as to provide a
                    # legal jump target and also ensure that the bytecode
                    # offsets are correct.
                    yield (offset, OPCODE_NOP, arg, i)
                    extended_arg = arg << 8 * ARG_LEN
                    offset = i
                    continue
            else:
                arg = None
                i += NO_ARG_LEN
                if PYVERSION in ((3, 12),):
                    # Python 3.12 introduced cache slots. We need to account for
                    # cache slots when we determine the offset of the next
                    # opcode. The number of cache slots is specific to each
                    # opcode and can be looked up in the _inline_cache_entries
                    # dictionary.
                    i += _inline_cache_entries[op] * INSTR_LEN
                elif PYVERSION in ((3, 9), (3, 10), (3, 11)):
                    pass
                else:
                    raise NotImplementedError(PYVERSION)

            extended_arg = 0
            yield (offset, op, arg, i)
            offset = i  # Mark inst offset at first extended

else:
    raise NotImplementedError(PYVERSION)


def _patched_opargs(bc_stream):
    """Patch the bytecode stream.

    - Adds a NOP bytecode at the start to avoid jump target being at the entry.
    """
    # Injected NOP
    yield (0, OPCODE_NOP, None, _FIXED_OFFSET)
    # Adjust bytecode offset for the rest of the stream
    for offset, opcode, arg, nextoffset in bc_stream:
        # If the opcode has an absolute jump target, adjust it.
        if opcode in JABS_OPS:
            arg += _FIXED_OFFSET
        yield offset + _FIXED_OFFSET, opcode, arg, nextoffset + _FIXED_OFFSET


class ByteCodeIter:
    def __init__(self, code):
        self.code = code
        self.iter = iter(_patched_opargs(_unpack_opargs(self.code.co_code)))

    def __iter__(self):
        return self

    def _fetch_opcode(self):
        return next(self.iter)

    def next(self):
        offset, opcode, arg, nextoffset = self._fetch_opcode()
        return offset, ByteCodeInst(offset=offset, opcode=opcode, arg=arg, nextoffset=nextoffset)

    __next__ = next

    def read_arg(self, size):
        buf = 0
        for i in range(size):
            _offset, byte = next(self.iter)
            buf |= byte << (8 * i)
        return buf


class _ByteCode:
    """
    The decoded bytecode of a function, and related information.
    """

    __slots__ = (
        "func_id",
        "co_names",
        "co_varnames",
        "co_consts",
        "co_cellvars",
        "co_freevars",
        "exception_entries",
        "table",
        "labels",
    )

    def __init__(self, func_id):
        code = func_id.code

        labels = set(x + _FIXED_OFFSET for x in dis.findlabels(code.co_code))
        labels.add(0)

        # A map of {offset: ByteCodeInst}
        table = OrderedDict(ByteCodeIter(code))
        self._compute_lineno(table, code)

        self.func_id = func_id
        self.co_names = code.co_names
        self.co_varnames = code.co_varnames
        self.co_consts = code.co_consts
        self.co_cellvars = code.co_cellvars
        self.co_freevars = code.co_freevars

        self.table = table
        self.labels = sorted(labels)

    @classmethod
    def _compute_lineno(cls, table, code):
        """
        Compute the line numbers for all bytecode instructions.
        """
        for offset, lineno in dis.findlinestarts(code):
            adj_offset = offset + _FIXED_OFFSET
            if adj_offset in table:
                table[adj_offset].lineno = lineno
        # Assign unfilled lineno
        # Start with first bytecode's lineno
        known = code.co_firstlineno
        for inst in table.values():
            if inst.lineno is not None and inst.lineno >= 0:
                known = inst.lineno
            else:
                inst.lineno = known
        return table

    def __iter__(self):
        return iter(self.table.values())

    def __getitem__(self, offset):
        return self.table[offset]

    def __contains__(self, offset):
        return offset in self.table

    def dump(self):
        def label_marker(i):
            if i[1].offset in self.labels:
                return ">"
            else:
                return " "

        return "\n".join(
            "%s %10s\t%s" % ((label_marker(i),) + i)
            for i in self.table.items()
            if i[1].opname != "CACHE"
        )

    @classmethod
    def _compute_used_globals(cls, func, table, co_consts, co_names):
        """
        Compute the globals used by the function with the given
        bytecode table.
        """
        d = {}
        globs = func.__globals__
        builtins = globs.get("__builtins__", utils.builtins)
        if isinstance(builtins, ModuleType):
            builtins = builtins.__dict__
        # Look for LOAD_GLOBALs in the bytecode
        for inst in table.values():
            if inst.opname == "LOAD_GLOBAL":
                name = co_names[_fix_LOAD_GLOBAL_arg(inst.arg)]
                if name not in d:
                    try:
                        value = globs[name]
                    except KeyError:
                        value = builtins[name]
                    d[name] = value
        # Add globals used by any nested code object
        for co in co_consts:
            if isinstance(co, CodeType):
                subtable = OrderedDict(ByteCodeIter(co))
                d.update(cls._compute_used_globals(func, subtable, co.co_consts, co.co_names))
        return d

    def get_used_globals(self):
        """
        Get a {name: value} map of the globals used by this code
        object and any nested code objects.
        """
        return self._compute_used_globals(
            self.func_id.func, self.table, self.co_consts, self.co_names
        )


def _fix_LOAD_GLOBAL_arg(arg):
    if PYVERSION in ((3, 11), (3, 12), (3, 13), (3, 14)):
        return arg >> 1
    elif PYVERSION in (
        (3, 9),
        (3, 10),
    ):
        return arg
    else:
        raise NotImplementedError(PYVERSION)


class ByteCodePy311(_ByteCode):
    def __init__(self, func_id):
        super().__init__(func_id)
        entries = dis.Bytecode(func_id.code).exception_entries
        self.exception_entries = tuple(map(self.fixup_eh, entries))

    @staticmethod
    def fixup_eh(ent):
        # Patch up the exception table offset
        # because we add a NOP in _patched_opargs
        out = dis._ExceptionTableEntry(
            start=ent.start + _FIXED_OFFSET,
            end=ent.end + _FIXED_OFFSET,
            target=ent.target + _FIXED_OFFSET,
            depth=ent.depth,
            lasti=ent.lasti,
        )
        return out

    def find_exception_entry(self, offset):
        """
        Returns the exception entry for the given instruction offset
        """
        candidates = [
            (ent.depth, ent) for ent in self.exception_entries if ent.start <= offset < ent.end
        ]
        if candidates:
            _, ent = max(candidates)
            return ent


class ByteCodePy312(ByteCodePy311):
    def __init__(self, func_id):
        super().__init__(func_id)

        # initialize lazy property
        self._ordered_offsets = None

        # Fixup offsets for all exception entries.
        entries = [self.fixup_eh(e) for e in dis.Bytecode(func_id.code).exception_entries]

        # Remove exceptions, innermost ones first
        # Can be done by using a stack
        entries = self.remove_build_list_swap_pattern(entries)

        # If this is a generator, we need to skip any exception table entries
        # that point to the exception handler with the highest offset.
        if func_id.is_generator:
            # Get the exception handler with the highest offset.
            max_exception_target = max([e.target for e in entries])
            # Remove any exception table entries that point to that exception
            # handler.
            entries = [e for e in entries if e.target != max_exception_target]

        self.exception_entries = tuple(entries)

    @property
    def ordered_offsets(self):
        if not self._ordered_offsets:
            # Get an ordered list of offsets.
            self._ordered_offsets = [o for o in self.table]
        return self._ordered_offsets

    def remove_build_list_swap_pattern(self, entries):
        """Find the following bytecode pattern:

        BUILD_{LIST, MAP, SET}
        SWAP(2)
        FOR_ITER
        ...
        END_FOR
        SWAP(2)

        This pattern indicates that a list/dict/set comprehension has
        been inlined. In this case we can skip the exception blocks
        entirely along with the dead exceptions that it points to.
        A pair of exception that sandwiches these exception will
        also be merged into a single exception.

        Update for Python 3.13, the ending of the pattern has a extra
        POP_TOP:

        ...
        END_FOR
        POP_TOP
        SWAP(2)

        Update for Python 3.13.1, there's now a GET_ITER before FOR_ITER.
        This patch the GET_ITER to NOP to minimize changes downstream
        (e.g. array-comprehension).
        """

        def pop_and_merge_exceptions(entries: list, entry_to_remove: _ExceptionTableEntry):
            lower_entry_idx = entries.index(entry_to_remove) - 1
            upper_entry_idx = entries.index(entry_to_remove) + 1

            # Merge the upper and lower exceptions if possible.
            if lower_entry_idx >= 0 and upper_entry_idx < len(entries):
                lower_entry = entries[lower_entry_idx]
                upper_entry = entries[upper_entry_idx]
                if lower_entry.target == upper_entry.target:
                    entries[lower_entry_idx] = _ExceptionTableEntry(
                        lower_entry.start,
                        upper_entry.end,
                        lower_entry.target,
                        lower_entry.depth,
                        upper_entry.lasti,
                    )
                    entries.remove(upper_entry)

            # Remove the exception entry.
            entries.remove(entry_to_remove)
            # Remove dead exceptions, if any, that the entry above may point to.
            entries = [e for e in entries if not e.start == entry_to_remove.target]
            return entries

        change_to_nop = set()
        work_remaining = True
        while work_remaining:
            # Temporarily set work_remaining to False, if we find a pattern
            # then work is not complete, hence we set it again to True.
            work_remaining = False
            current_nop_fixes = set()
            for entry in entries.copy():
                # Check start of pattern, three instructions.
                # Work out the index of the instruction.
                index = self.ordered_offsets.index(entry.start)
                # If there is a BUILD_{LIST, MAP, SET} instruction at this
                # location.
                curr_inst = self.table[self.ordered_offsets[index]]
                if curr_inst.opname not in (
                    "BUILD_LIST",
                    "BUILD_MAP",
                    "BUILD_SET",
                ):
                    continue
                # Check if the BUILD_{LIST, MAP, SET} instruction is followed
                # by a SWAP(2).
                next_inst = self.table[self.ordered_offsets[index + 1]]
                if not next_inst.opname == "SWAP" and next_inst.arg == 2:
                    continue
                next_inst = self.table[self.ordered_offsets[index + 2]]
                # Check if the SWAP is followed by a FOR_ITER
                # BUT Python3.13.1 introduced an extra GET_ITER.
                # If we see a GET_ITER here, check if the next thing is a
                # FOR_ITER.
                if next_inst.opname == "GET_ITER":
                    # Add the inst to potentially be replaced to NOP
                    current_nop_fixes.add(next_inst)
                    # Loop up next instruction.
                    next_inst = self.table[self.ordered_offsets[index + 3]]

                if not next_inst.opname == "FOR_ITER":
                    continue

                if PYVERSION in ((3, 13), (3, 14)):
                    # Check end of pattern, two instructions.
                    # Check for the corresponding END_FOR, exception table end
                    # is non-inclusive, so subtract one.
                    index = self.ordered_offsets.index(entry.end)
                    curr_inst = self.table[self.ordered_offsets[index - 2]]
                    if not curr_inst.opname == "END_FOR":
                        continue
                    next_inst = self.table[self.ordered_offsets[index - 1]]
                    if PYVERSION in ((3, 13),):
                        if not next_inst.opname == "POP_TOP":
                            continue
                    elif PYVERSION in ((3, 14),):
                        if not next_inst.opname == "POP_ITER":
                            continue
                    else:
                        raise NotImplementedError(PYVERSION)
                    # END_FOR must be followed by SWAP(2)
                    next_inst = self.table[self.ordered_offsets[index]]
                    if not next_inst.opname == "SWAP" and next_inst.arg == 2:
                        continue
                elif PYVERSION in ((3, 9), (3, 10), (3, 11), (3, 12)):
                    # Check end of pattern, two instructions.
                    # Check for the corresponding END_FOR, exception table end
                    # is non-inclusive, so subtract one.
                    index = self.ordered_offsets.index(entry.end)
                    curr_inst = self.table[self.ordered_offsets[index - 1]]
                    if not curr_inst.opname == "END_FOR":
                        continue
                    # END_FOR must be followed by SWAP(2)
                    next_inst = self.table[self.ordered_offsets[index]]
                    if not next_inst.opname == "SWAP" and next_inst.arg == 2:
                        continue
                else:
                    raise NotImplementedError(PYVERSION)
                # If all conditions are met that means this exception entry
                # is for a list/dict/set comprehension and can be removed.
                # Also if there exist exception entries above and below this
                # entry pointing to the same target. those can be merged into
                # a single bigger exception block.
                entries = pop_and_merge_exceptions(entries, entry)
                work_remaining = True

                # Commit NOP fixes since we confirmed the suspects belong to
                # a comprehension code.
                change_to_nop |= current_nop_fixes

        # Complete fixes to NOPs
        for inst in change_to_nop:
            self.table[inst.offset] = ByteCodeInst(inst.offset, dis.opmap["NOP"], None, inst.next)
        return entries


if PYVERSION == (3, 11):
    ByteCode = ByteCodePy311
elif PYVERSION in (
    (3, 12),
    (3, 13),
    (3, 14),
):
    ByteCode = ByteCodePy312
elif PYVERSION < (3, 11):
    ByteCode = _ByteCode
else:
    raise NotImplementedError(PYVERSION)


class FunctionIdentity(serialize.ReduceMixin):
    """
    A function's identity and metadata.

    Note this typically represents a function whose bytecode is
    being compiled, not necessarily the top-level user function
    (the two might be distinct).
    """

    _unique_ids = _LockedCounter(1)

    @classmethod
    def from_function(cls, pyfunc):
        """
        Create the FunctionIdentity of the given function.
        """
        func = get_function_object(pyfunc)
        code = get_code_object(func)
        pysig = utils.pysignature(func)
        if not code:
            raise errors.ByteCodeSupportError("%s does not provide its bytecode" % func)

        try:
            func_qualname = func.__qualname__
        except AttributeError:
            func_qualname = func.__name__

        self = cls()
        self.func = func
        self.func_qualname = func_qualname
        self.func_name = func_qualname.split(".")[-1]
        self.code = code
        self.module = inspect.getmodule(func)
        self.modname = utils._dynamic_modname if self.module is None else self.module.__name__
        self.is_generator = inspect.isgeneratorfunction(func)
        self.pysig = pysig
        self.filename = code.co_filename
        self.firstlineno = code.co_firstlineno
        self.arg_count = len(pysig.parameters)
        self.arg_names = list(pysig.parameters)

        # Even the same function definition can be compiled into
        # several different function objects with distinct closure
        # variables, so we make sure to disambiguate using an unique id.
        uid = next(cls._unique_ids)
        self.unique_name = "{}${}".format(self.func_qualname, uid)
        self.unique_id = uid

        return self

    def derive(self):
        """Copy the object and increment the unique counter."""
        return self.from_function(self.func)

    def _reduce_states(self):
        """
        NOTE: part of ReduceMixin protocol
        """
        return dict(pyfunc=self.func)

    @classmethod
    def _rebuild(cls, pyfunc):
        """
        NOTE: part of ReduceMixin protocol
        """
        return cls.from_function(pyfunc)



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AbstractTemplate,
    AttributeTemplate,
    Registry,
    signature,
)
from numba_cuda_mlir import types
from numba_cuda_mlir.type_defs.vector_types import VectorType
from numba_cuda_mlir.errors import ForceLiteralArg

registry = Registry()


def _extract_shape(shape_arg):
    """Extract shape from literal argument."""
    if isinstance(shape_arg, types.IntegerLiteral):
        return (shape_arg.literal_value,)
    elif isinstance(shape_arg, (types.Tuple, types.UniTuple)):
        if hasattr(shape_arg, "types"):
            shape = []
            for t in shape_arg.types:
                if isinstance(t, types.IntegerLiteral):
                    shape.append(t.literal_value)
                else:
                    return None
            return tuple(shape)
    return None


@registry.register
class VectorLoadTemplate(AbstractTemplate):
    from numba_cuda_mlir.cuda import vector

    key = vector.load

    def generic(self, args, kws):
        # Handle both 3 args (no alignment) and 4 args (with alignment)
        if len(args) not in (3, 4):
            return None

        array, index, shape = args[0], args[1], args[2]
        alignment = args[3] if len(args) == 4 else None

        if not isinstance(array, (types.Array, types.CPointer)):
            return None

        # CPointer only supports 1D (scalar) index
        if isinstance(array, types.CPointer) and isinstance(index, (types.Tuple, types.UniTuple)):
            raise TypeError(
                "vector.load with CPointer only supports scalar index, not tuple. "
                "Linearize the index before passing to vector.load."
            )

        vec_shape = _extract_shape(shape)
        if vec_shape is None:
            if isinstance(shape, types.Integer):
                raise ForceLiteralArg({2})
            return None

        # If alignment is provided, check its type
        if alignment is not None:
            if isinstance(alignment, types.NoneType):
                # Explicit None passed - treat as unaligned, fall through
                pass
            elif isinstance(alignment, types.IntegerLiteral):
                # Aligned path - alignment must be a compile-time constant
                return signature(VectorType(array.dtype, vec_shape), array, index, shape, alignment)
            elif isinstance(alignment, types.Integer):
                # Force user to provide a literal for alignment
                raise ForceLiteralArg({3})
            else:
                # Invalid alignment type
                return None

        # No alignment provided or explicit None - unaligned path
        if isinstance(array, types.CPointer):
            raise TypeError(
                "vector.load with CPointer requires an explicit alignment argument. "
                "Use vector.load(ptr, index, shape, alignment=N)."
            )
        return signature(VectorType(array.dtype, vec_shape), array, index, shape)


@registry.register
class VectorStoreTemplate(AbstractTemplate):
    from numba_cuda_mlir.cuda import vector

    key = vector.store

    def generic(self, args, kws):
        # Handle both 3 args (no alignment) and 4 args (with alignment)
        if len(args) not in (3, 4):
            return None

        array, index, vec = args[0], args[1], args[2]
        alignment = args[3] if len(args) == 4 else None

        if not isinstance(array, (types.Array, types.CPointer)):
            return None
        if not isinstance(vec, VectorType):
            return None

        # CPointer only supports 1D (scalar) index
        if isinstance(array, types.CPointer) and isinstance(index, (types.Tuple, types.UniTuple)):
            raise TypeError(
                "vector.store with CPointer only supports scalar index, not tuple. "
                "Linearize the index before passing to vector.store."
            )

        # If alignment is provided, check its type
        if alignment is not None:
            if isinstance(alignment, types.NoneType):
                # Explicit None passed - treat as unaligned, fall through
                pass
            elif isinstance(alignment, types.IntegerLiteral):
                # Aligned path - alignment must be a compile-time constant
                return signature(types.none, array, index, vec, alignment)
            elif isinstance(alignment, types.Integer):
                # Force user to provide a literal for alignment
                raise ForceLiteralArg({3})
            else:
                # Invalid alignment type
                return None

        # No alignment provided or explicit None - unaligned path
        if isinstance(array, types.CPointer):
            raise TypeError(
                "vector.store with CPointer requires an explicit alignment argument. "
                "Use vector.store(ptr, index, vec, alignment=N)."
            )
        return signature(types.none, array, index, vec)


@registry.register_attr
class VectorModuleTemplate(AttributeTemplate):
    from numba_cuda_mlir.cuda import vector as vector_module

    key = types.Module(vector_module)

    def resolve_load(self, mod):
        return types.Function(VectorLoadTemplate)

    def resolve_store(self, mod):
        return types.Function(VectorStoreTemplate)




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class ApplyNVGPUToNVVMConversionPatternsOp(_ods_ir.OpView):
  r"""
  Collects patterns that convert NVGPU dialect ops to NVVM dialect ops. These
  patterns require an "LLVMTypeConverter".
  """

  OPERATION_NAME = "transform.apply_conversion_patterns.nvgpu.nvgpu_to_nvvm"

  _ODS_REGIONS = (0, True)

  def __init__(self, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

@_ods_cext.register_op_adaptor(ApplyNVGPUToNVVMConversionPatternsOp)
class ApplyNVGPUToNVVMConversionPatternsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.apply_conversion_patterns.nvgpu.nvgpu_to_nvvm"

def apply_conversion_patterns_nvgpu_nvgpu_to_nvvm(*, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> ApplyNVGPUToNVVMConversionPatternsOp:
  return ApplyNVGPUToNVVMConversionPatternsOp(loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class CreateAsyncGroupsOp(_ods_ir.OpView):
  r"""
  Look for global to shared memory copies within the targeted op in the form
  of vector transfer ops and convert them to async copies when possible.
  Consecutive copies are put into the same group. A "wait" operation is
  inserted right at the of end the group.
  
  `bypass_l1` specifies whether `bypassL1` attributes should be added to
  the async copies. `bypass_l1` is a compiler hint: only 16 byte transfers
  can bypass the L1 cache, so this attribute is not set for any other transfer
  sizes.
  
  #### Return modes
  
  This op consumes the `target` handle and produces the `result` handle, which
  is mapped to the same payload operations as the `target` handle. The op
  modifies the payload.
  """

  OPERATION_NAME = "transform.nvgpu.create_async_groups"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, target: _ods_ir.Value, *, bypass_l1: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if bool(bypass_l1): attributes["bypass_l1"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def bypass_l1(self) -> bool:
    return "bypass_l1" in self.operation.attributes

  @bypass_l1.setter
  def bypass_l1(self, value):
    if bool(value):
      self.operation.attributes["bypass_l1"] = _ods_ir.UnitAttr.get()
    elif "bypass_l1" in self.operation.attributes:
      del self.operation.attributes["bypass_l1"]

  @bypass_l1.deleter
  def bypass_l1(self):
    del self.operation.attributes["bypass_l1"]

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(CreateAsyncGroupsOp)
class CreateAsyncGroupsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.nvgpu.create_async_groups"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def bypass_l1(self) -> bool:
    return "bypass_l1" in self.attributes

def nvgpu_create_async_groups(result: _ods_ir.Type, target: _ods_ir.Value, *, bypass_l1: _Optional[bool] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return CreateAsyncGroupsOp(result=result, target=target, bypass_l1=bypass_l1, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class PipelineSharedMemoryCopiesOp(_ods_ir.OpView):
  r"""
  Applies software pipelining to a given scf.for loop. The pipelining
  strategy will look for a load into shared memory and pipeline it to overlap
  it with the rest of the loop.
  
  NOTE: It is user responsibility to ensure that there are no dependency
  between `depth` iterations of the loop by using multi-buffering. It is
  also user responsibility to ensure a sufficient amount of shared memory
  is allocated to cover eventual writes by `depth-1` speculative
  iterations.
  
  `depth` will indicate how many stages the software pipeline should have.
  `peel_epilogue` allows to force the epilogue to be peeled out instead of
  potentially using predicated operations for the epilogue phase.
  
  #### Return modes
  
  Consumes the operand handle and produces a result handle pointing to the
  loop, which may or may not have been pipelined. Produces a definite failure
  if the loop pipeliner mutated the IR before failing to pipeline, in
  particular if `peel_epilogue` is not set and the loop body doesn't support
  predication. If failure propagation mode is set to "propagate", produces a
  silenceable failure when pipelining preconditions, e.g., loop bound being
  static, are not met or when the loop wasn't pipelined because due to the
  lack of loads into shared memory. If the failure propagation mode is set
  to "suppress" (default), succeeds in these case and associates the result
  handle with the original loop.
  
  TODO: the shared memory part and behavior specific to NVGPU should be
  made orthogonal to pipelining so that `transform.loop.pipeline` becomes
  usable here.
  """

  OPERATION_NAME = "transform.nvgpu.pipeline_shared_memory_copies"

  _ODS_REGIONS = (0, True)

  def __init__(self, result: _ods_ir.Type, for_op: _ods_ir.Value, depth: _Union[int, _ods_ir.IntegerAttr], *, peel_epilogue: _Optional[bool] = None, failure_propagation_mode: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(for_op)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["depth"] = (depth if (
    isinstance(depth, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('I64Attr')) else
      _ods_ir.AttrBuilder.get('I64Attr')(depth, context=_ods_context))
    if bool(peel_epilogue): attributes["peel_epilogue"] = _ods_ir.UnitAttr.get(
      _ods_get_default_loc_context(loc))
    if failure_propagation_mode is not None: attributes["failure_propagation_mode"] = (failure_propagation_mode if (
        isinstance(failure_propagation_mode, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('FailurePropagationMode')) else
          _ods_ir.AttrBuilder.get('FailurePropagationMode')(failure_propagation_mode, context=_ods_context))
    results = []
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def for_op(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def depth(self) -> _ods_ir.IntegerAttr:
    return self.operation.attributes["depth"]

  @depth.setter
  def depth(self, value: _ods_ir.IntegerAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["depth"] = value

  @builtins.property
  def peel_epilogue(self) -> bool:
    return "peel_epilogue" in self.operation.attributes

  @peel_epilogue.setter
  def peel_epilogue(self, value):
    if bool(value):
      self.operation.attributes["peel_epilogue"] = _ods_ir.UnitAttr.get()
    elif "peel_epilogue" in self.operation.attributes:
      del self.operation.attributes["peel_epilogue"]

  @peel_epilogue.deleter
  def peel_epilogue(self):
    del self.operation.attributes["peel_epilogue"]

  @builtins.property
  def failure_propagation_mode(self) -> _ods_ir.Attribute:
    return self.operation.attributes["failure_propagation_mode"]

  @failure_propagation_mode.setter
  def failure_propagation_mode(self, value: _ods_ir.Attribute):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["failure_propagation_mode"] = value

  @builtins.property
  def result(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(PipelineSharedMemoryCopiesOp)
class PipelineSharedMemoryCopiesOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.nvgpu.pipeline_shared_memory_copies"

  @builtins.property
  def for_op(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def depth(self) -> _ods_ir.IntegerAttr:
    return self.attributes["depth"]

  @builtins.property
  def peel_epilogue(self) -> bool:
    return "peel_epilogue" in self.attributes

  @builtins.property
  def failure_propagation_mode(self) -> _ods_ir.Attribute:
    return self.attributes["failure_propagation_mode"]

def nvgpu_pipeline_shared_memory_copies(result: _ods_ir.Type, for_op: _ods_ir.Value, depth: _Union[int, _ods_ir.IntegerAttr], *, peel_epilogue: _Optional[bool] = None, failure_propagation_mode: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return PipelineSharedMemoryCopiesOp(result=result, for_op=for_op, depth=depth, peel_epilogue=peel_epilogue, failure_propagation_mode=failure_propagation_mode, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class RewriteCopyAsTmaOp(_ods_ir.OpView):
  r"""
  Rewrite a copy operation on memref to tma operations that transit through
  shared memory.
  """

  OPERATION_NAME = "transform.nvgpu.rewrite_copy_as_tma"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RewriteCopyAsTmaOp)
class RewriteCopyAsTmaOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.nvgpu.rewrite_copy_as_tma"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def nvgpu_rewrite_copy_as_tma(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RewriteCopyAsTmaOp:
  return RewriteCopyAsTmaOp(target=target, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class RewriteMatmulAsMmaSyncOp(_ods_ir.OpView):
  r"""
  Rewrite a matmul operation on memref to an mma.sync operation on vectors.
  
  Memory copies with the required access patterns are automatically inserted.
  Operations that do not have a 1-1 mapping to mma.sync operations are left
  unchanged.
  """

  OPERATION_NAME = "transform.nvgpu.rewrite_matmul_as_mma_sync"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(RewriteMatmulAsMmaSyncOp)
class RewriteMatmulAsMmaSyncOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.nvgpu.rewrite_matmul_as_mma_sync"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def nvgpu_rewrite_matmul_as_mma_sync(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> RewriteMatmulAsMmaSyncOp:
  return RewriteMatmulAsMmaSyncOp(target=target, loc=loc, ip=ip)




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class FailurePropagationMode(IntEnum):
    """Silenceable error propagation policy"""

    Propagate = 1
    Suppress = 2

    def __str__(self):
        if self is FailurePropagationMode.Propagate:
            return "propagate"
        if self is FailurePropagationMode.Suppress:
            return "suppress"
        raise ValueError("Unknown FailurePropagationMode enum entry.")



@register_attribute_builder("FailurePropagationMode", allow_existing=True)
def _failurepropagationmode(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))

class MatchCmpIPredicate(IntEnum):
    """allowed 32-bit signless integer cases: 0, 1, 2, 3, 4, 5"""

    eq = 0
    ne = 1
    lt = 2
    le = 3
    gt = 4
    ge = 5

    def __str__(self):
        if self is MatchCmpIPredicate.eq:
            return "eq"
        if self is MatchCmpIPredicate.ne:
            return "ne"
        if self is MatchCmpIPredicate.lt:
            return "lt"
        if self is MatchCmpIPredicate.le:
            return "le"
        if self is MatchCmpIPredicate.gt:
            return "gt"
        if self is MatchCmpIPredicate.ge:
            return "ge"
        raise ValueError("Unknown MatchCmpIPredicate enum entry.")



@register_attribute_builder("MatchCmpIPredicateAttr", allow_existing=True)
def _matchcmpipredicateattr(x, context):
    return _ods_ir.IntegerAttr.get(_ods_ir.IntegerType.get_signless(32, context=context), int(x))




#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""DSL for constructing affine expressions and maps.

These python wrappers allow construction of affine expressions in a more
pythonic fashion that is later instantiated as an IR AffineExpr. Separating the
AST from construction of the map allows for manipulations of symbols and dims
beyond the scope of one expression.

Affine expression construction:
  >>> with _ir.Context():
  ...   s = AffineBuildState()
  ...   (S.K + S.M).build(s)
  ...   (S.K * S.M).build(s)
  ...   (S.K // S.M).build(s)
  ...   (S.K / S.M).build(s)
  ...   (S.K % 4).build(s)
  ...   (D.i + D.j * 4).build(s)
  ...   s
  AffineExpr(s0 + s1)
  AffineExpr(s0 * s1)
  AffineExpr(s0 floordiv s1)
  AffineExpr(s0 ceildiv s1)
  AffineExpr(s0 mod 4)
  AffineExpr(d0 + d1 * 4)
  AffineBuildState<
    symbols={'K': 0, 'M': 1}
    dims={'i': 0, 'j': 1}>

In the DSL, dimensions and symbols are name-uniqued instances of DimDef and
SymbolDef. There are shortcut "expando" instances that will create a
corresponding DimDef/SymbolDef upon accessing an attribute:

Referencing a named dimension:

  >>> D.i
  Dim(i)
  >>> D.a is D.b
  False
  >>> D.a is D.a
  True

Referencing a named symbol:

  >>> S.foobar
  Symbol(foobar)
  >>> S.a is S.b
  False
  >>> S.a is S.a
  True
"""

from typing import Callable, Dict, Optional, Tuple, Union

from ..... import ir as _ir

__all__ = [
    "AffineBuildState",
    "AffineExprDef",
    "D",
    "DimDef",
    "S",
    "SymbolDef",
]


class AffineBuildState:
    """Internal state for the AffineExprDef._create impls.

    Note that a "local" AffineBuildState can be created relative to a "global"
    AffineBuildState. In that case, any affine expressions built will inherit
    symbol and dim bindings from the global state and will update both as new
    ones are discovered. This allows for building expressions across contexts
    which share a common symbol and dim space.
    """

    def __init__(
        self,
        *,
        global_state: "AffineBuildState" = None,
        allow_new_symbols: bool = True,
        allow_new_dims: bool = True,
    ):
        if not global_state:
            self.all_symbols = dict()  # type: Dict[str, int]
            self.all_dims = dict()  # type: Dict[str, int]
        else:
            # Alias the global dict.
            self.all_symbols = global_state.all_symbols
            self.all_dims = global_state.all_dims

        # Map of symbols and dims in the current build.
        self.local_symbols = dict()  # type: Dict[str, int]
        self.local_dims = dict()  # type: Dict[str, int]
        self.allow_new_symbols = allow_new_symbols
        self.allow_new_dims = allow_new_dims

    def get_dim(self, dimname: str) -> int:
        """Gets the dim position given a name."""
        pos = self.all_dims.get(dimname)
        if pos is None:
            if not self.allow_new_dims:
                raise ValueError(
                    f"New dimensions not allowed in the current affine expression: "
                    f"Requested '{dimname}', Availble: {self.all_dims}"
                )
            pos = len(self.all_dims)
            self.all_dims[dimname] = pos
        self.local_dims[dimname] = pos
        return pos

    def get_symbol(self, symname: str) -> int:
        """Geta a symbol position given a name."""
        pos = self.all_symbols.get(symname)
        if pos is None:
            if not self.allow_new_symbols:
                raise ValueError(
                    f"New symbols not allowed in the current affine expression: "
                    f"Requested '{symname}', Availble: {self.all_symbols}"
                )
            pos = len(self.all_symbols)
            self.all_symbols[symname] = pos
        self.local_symbols[symname] = pos
        return pos

    @property
    def local_dim_count(self) -> int:
        return len(self.local_dims)

    @property
    def local_symbol_count(self) -> int:
        return len(self.local_symbols)

    @property
    def dim_count(self) -> int:
        return len(self.all_dims)

    @property
    def symbol_count(self) -> int:
        return len(self.all_symbols)

    def __repr__(self):
        lines = [f"AffineBuildState<"]
        lines.append(f"  symbols={self.local_symbols}")
        lines.append(f"  dims={self.local_dims}>")
        return "\n".join(lines)


class AffineExprDef:
    """Base class for an affine expression being defined."""

    def build(self, state: Optional[AffineBuildState] = None) -> _ir.AffineExpr:
        """Builds the corresponding _ir.AffineExpr from the definitions."""
        state = AffineBuildState() if state is None else state
        expr = self._create(state)
        return expr

    def _create(self, state: AffineBuildState) -> _ir.AffineExpr:
        raise NotImplementedError()

    @staticmethod
    def coerce_from(py_value):
        if isinstance(py_value, int):
            return AffineConstantExpr(py_value)
        assert isinstance(py_value, AffineExprDef)
        return py_value

    def visit_affine_exprs(self, callback):
        """Visits all AffineExprDefs including self."""
        callback(self)

    def __add__(lhs, rhs):
        rhs = AffineExprDef.coerce_from(rhs)
        return AffineBinaryExprDef(_ir.AffineAddExpr, lhs, rhs)

    def __mul__(lhs, rhs):
        rhs = AffineExprDef.coerce_from(rhs)
        return AffineBinaryExprDef(_ir.AffineMulExpr, lhs, rhs)

    def __mod__(lhs, rhs):
        rhs = AffineExprDef.coerce_from(rhs)
        return AffineBinaryExprDef(_ir.AffineModExpr, lhs, rhs)

    def __floordiv__(lhs, rhs):
        rhs = AffineExprDef.coerce_from(rhs)
        return AffineBinaryExprDef(_ir.AffineFloorDivExpr, lhs, rhs)

    def __truediv__(lhs, rhs):
        # TODO: Not really a ceil div - taking liberties for the DSL.
        rhs = AffineExprDef.coerce_from(rhs)
        return AffineBinaryExprDef(_ir.AffineCeilDivExpr, lhs, rhs)


class AffineConstantExpr(AffineExprDef):
    """An affine constant being defined."""

    def __init__(self, value: int):
        assert isinstance(value, int)
        self.value = value

    def _create(self, state: AffineBuildState) -> _ir.AffineExpr:
        return _ir.AffineConstantExpr.get(self.value)

    def __repr__(self):
        return f"Const({self.value})"


class AffineBinaryExprDef(AffineExprDef):
    """An affine binary expression being defined."""

    def __init__(self, ir_ctor, lhs: AffineExprDef, rhs: AffineExprDef):
        self.ir_ctor = ir_ctor
        self.lhs = lhs
        self.rhs = rhs

    def _create(self, state: AffineBuildState) -> _ir.AffineExpr:
        return self.ir_ctor.get(self.lhs._create(state), self.rhs._create(state))

    def visit_affine_exprs(self, callback):
        """Visits all AffineExprDefs including self."""
        super().visit_affine_exprs(callback)
        self.lhs.visit_affine_exprs(callback)
        self.rhs.visit_affine_exprs(callback)

    def __repr__(self):
        return f"{self.ir_ctor.__name__}({repr(self.lhs)}, {repr(self.rhs)})"


class DimDef(AffineExprDef):
    """Represents a named dimension."""

    ALL_DIMS = dict()  # type: Dict[str, "DimDef"]

    def __new__(cls, dimname: str):
        existing = cls.ALL_DIMS.get(dimname)
        if existing is not None:
            return existing
        new = super().__new__(cls)
        new.dimname = dimname
        cls.ALL_DIMS[dimname] = new
        return new

    def __repr__(self):
        return f"Dim({self.dimname})"

    def _create(self, state: AffineBuildState) -> _ir.AffineExpr:
        pos = state.get_dim(self.dimname)
        return _ir.AffineDimExpr.get(position=pos)

    @classmethod
    def create_expando(cls):
        """Create an expando class that creates unique symbols based on attr access."""

        class ExpandoDims:
            def __getattr__(self, n):
                return cls(n)

        return ExpandoDims()


class SymbolDef(AffineExprDef):
    """Represents a named symbol.

    >>> s1 = SymbolDef("s1")
    >>> s1
    Symbol(s1)
    >>> s2 = SymbolDef("s2")
    >>> s1 is s2
    False
    >>> s1 is SymbolDef("s1")
    True
    """

    ALL_SYMBOLS = dict()  # type: Dict[str, "SymbolDef"]

    def __new__(cls, symname: str):
        existing = cls.ALL_SYMBOLS.get(symname)
        if existing is not None:
            return existing
        new = super().__new__(cls)
        new.symname = symname
        cls.ALL_SYMBOLS[symname] = new
        return new

    def __repr__(self):
        return f"Symbol({self.symname})"

    def _create(self, state: AffineBuildState) -> _ir.AffineExpr:
        pos = state.get_symbol(self.symname)
        return _ir.AffineSymbolExpr.get(position=pos)

    @classmethod
    def create_expando(cls):
        """Create an expando class that creates unique symbols based on attr access."""

        class ExpandoSymbols:
            def __getattr__(self, n):
                return cls(n)

        return ExpandoSymbols()


# Global accessor for on-demand dims and symbols.
D = DimDef.create_expando()
S = SymbolDef.create_expando()




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class ConstrainParamsOp(_ods_ir.OpView):
  r"""
  Allows expressing constraints on params using the SMT dialect.
  
  Each Transform-dialect param provided as an operand has a corresponding
  argument of SMT-type in the region. The SMT-Dialect ops in the region use
  these params-as-SMT-vars as operands, thereby expressing relevant
  constraints on their allowed values.
  
  Computations w.r.t. passed-in params can also be expressed through the
  region's SMT-ops. Namely, the constraints express relationships to other
  SMT-variables which can then be yielded from the region (with `smt.yield`).
  
  The semantics of this op is that all the ops in the region together express
  a constraint on the params-interpreted-as-smt-vars. The op fails in case the
  expressed constraint is not satisfiable per SMTLIB semantics. Otherwise the
  op succeeds and any one satisfying assignment is used to map the
  SMT-variables yielded in the region to `transform.param`s.
  
  ---
  
  TODO: currently the operational semantics per the Transform interpreter is
  to always fail. The intention is build out support for hooking in your own
  operational semantics so you can invoke your favourite solver to determine
  satisfiability of the corresponding constraint problem.
  """

  OPERATION_NAME = "transform.smt.constrain_params"

  _ODS_REGIONS = (1, True)

  def __init__(self, results_: _Sequence[_ods_ir.Type], params: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(params))
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.extend(results_)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def params(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operation.operands) - 1 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def results_(self) -> _ods_ir.OpResultList:
    _ods_variadic_group_length = len(self.operation.results) - 1 + 1
    return self.operation.results[0:0 + _ods_variadic_group_length]

  @builtins.property
  def body(self) -> _ods_ir.Region:
    return self.regions[0]

@_ods_cext.register_op_adaptor(ConstrainParamsOp)
class ConstrainParamsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.smt.constrain_params"

  @builtins.property
  def params(self) -> _ods_ir.OpOperandList:
    _ods_variadic_group_length = len(self.operands) - 1 + 1
    return self.operands[0:0 + _ods_variadic_group_length]

def smt_constrain_params(results_: _Sequence[_ods_ir.Type], params: _Sequence[_ods_ir.Value], *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _Union[_ods_ir.OpResult, _ods_ir.OpResultList, ConstrainParamsOp]:
  op = ConstrainParamsOp(results_=results_, params=params, loc=loc, ip=ip); results = op.results
  return results if len(results) > 1 else (results[0] if len(results) == 1 else op)



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
Implement the cmath module functions.

The ``@lower`` codegen that used to live here (cmath.isnan/isinf/isfinite,
exp, log, sqrt, cos, sin, tan, acos, asin, atan, asinh) built llvmlite IR and
was installed into the target context, but it is filtered out on the MLIR path
(numba_cuda_mlir.lowering.cmath lowers these). Only the ``@overload``
implementations remain: those provide pure-Python implementations that the MLIR
path type-infers and lowers normally.
"""

import cmath
import math

from numba_cuda_mlir.numba_cuda.core.imputils import Registry
from numba_cuda_mlir.numba_cuda import types
from numba_cuda_mlir.numba_cuda.cpython import mathimpl
from numba_cuda_mlir.numba_cuda.extending import overload


registry = Registry("cmathimpl")
lower = registry.lower


@overload(cmath.rect)
def impl_cmath_rect(r, phi):
    if all([isinstance(typ, types.Float) for typ in [r, phi]]):

        def impl(r, phi):
            if not math.isfinite(phi):
                if not r:
                    # cmath.rect(0, phi={inf, nan}) = 0
                    return abs(r)
                if math.isinf(r):
                    # cmath.rect(inf, phi={inf, nan}) = inf + j phi
                    return complex(r, phi)
            real = math.cos(phi)
            imag = math.sin(phi)
            if real == 0.0 and math.isinf(r):
                # 0 * inf would return NaN, we want to keep 0 but xor the sign
                real /= r
            else:
                real *= r
            if imag == 0.0 and math.isinf(r):
                # ditto
                imag /= r
            else:
                imag *= r
            return complex(real, imag)

        return impl


@overload(cmath.log10)
def impl_cmath_log10(z):
    if not isinstance(z, types.Complex):
        return

    LN_10 = 2.302585092994045684

    def log10_impl(z):
        """cmath.log10(z)"""
        z = cmath.log(z)
        # This formula gives better results on +/-inf than cmath.log(z, 10)
        # See http://bugs.python.org/issue22544
        return complex(z.real / LN_10, z.imag / LN_10)

    return log10_impl


@overload(cmath.phase)
def phase_impl(x):
    """cmath.phase(x + y j)"""

    if not isinstance(x, types.Complex):
        return

    def impl(x):
        return math.atan2(x.imag, x.real)

    return impl


@overload(cmath.polar)
def polar_impl(x):
    if not isinstance(x, types.Complex):
        return

    def impl(x):
        r, i = x.real, x.imag
        return math.hypot(r, i), math.atan2(i, r)

    return impl


@overload(cmath.cosh)
def impl_cmath_cosh(z):
    if not isinstance(z, types.Complex):
        return

    def cosh_impl(z):
        """cmath.cosh(z)"""
        x = z.real
        y = z.imag
        if math.isinf(x):
            if math.isnan(y):
                # x = +inf, y = NaN => cmath.cosh(x + y j) = inf + Nan * j
                real = abs(x)
                imag = y
            elif y == 0.0:
                # x = +inf, y = 0 => cmath.cosh(x + y j) = inf + 0j
                real = abs(x)
                imag = y
            else:
                real = math.copysign(x, math.cos(y))
                imag = math.copysign(x, math.sin(y))
            if x < 0.0:
                # x = -inf => negate imaginary part of result
                imag = -imag
            return complex(real, imag)
        return complex(math.cos(y) * math.cosh(x), math.sin(y) * math.sinh(x))

    return cosh_impl


@overload(cmath.sinh)
def impl_cmath_sinh(z):
    if not isinstance(z, types.Complex):
        return

    def sinh_impl(z):
        """cmath.sinh(z)"""
        x = z.real
        y = z.imag
        if math.isinf(x):
            if math.isnan(y):
                # x = +/-inf, y = NaN => cmath.sinh(x + y j) = x + NaN * j
                real = x
                imag = y
            else:
                real = math.cos(y)
                imag = math.sin(y)
                if real != 0.0:
                    real *= x
                if imag != 0.0:
                    imag *= abs(x)
            return complex(real, imag)
        return complex(math.cos(y) * math.sinh(x), math.sin(y) * math.cosh(x))

    return sinh_impl


@overload(cmath.tanh)
def impl_cmath_tanh(z):
    if not isinstance(z, types.Complex):
        return

    def tanh_impl(z):
        """cmath.tanh(z)"""
        x = z.real
        y = z.imag
        if math.isinf(x):
            real = math.copysign(1.0, x)
            if math.isinf(y):
                imag = 0.0
            else:
                imag = math.copysign(0.0, math.sin(2.0 * y))
            return complex(real, imag)
        # This is CPython's algorithm (see c_tanh() in cmathmodule.c).
        # XXX how to force float constants into single precision?
        tx = math.tanh(x)
        ty = math.tan(y)
        cx = 1.0 / math.cosh(x)
        txty = tx * ty
        denom = 1.0 + txty * txty
        return complex(tx * (1.0 + ty * ty) / denom, ((ty / denom) * cx) * cx)

    return tanh_impl


@overload(cmath.acosh)
def impl_cmath_acosh(z):
    if not isinstance(z, types.Complex):
        return

    LN_4 = math.log(4)
    THRES = mathimpl.FLT_MAX / 4

    def acosh_impl(z):
        """cmath.acosh(z)"""
        # CPython's algorithm (see c_acosh() in cmathmodule.c)
        if abs(z.real) > THRES or abs(z.imag) > THRES:
            # Avoid unnecessary overflow for large arguments
            # (also handles infinities gracefully)
            real = math.log(math.hypot(z.real * 0.5, z.imag * 0.5)) + LN_4
            imag = math.atan2(z.imag, z.real)
            return complex(real, imag)
        else:
            s1 = cmath.sqrt(complex(z.real - 1.0, z.imag))
            s2 = cmath.sqrt(complex(z.real + 1.0, z.imag))
            real = math.asinh(s1.real * s2.real + s1.imag * s2.imag)
            imag = 2.0 * math.atan2(s1.imag, s2.real)
            return complex(real, imag)
        # Condensed formula (NumPy)
        # return cmath.log(z + cmath.sqrt(z + 1.) * cmath.sqrt(z - 1.))

    return acosh_impl



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from numba_cuda_mlir.numba_cuda import typing, types
from dataclasses import dataclass
from numba_cuda_mlir.numba_cuda.libdevicefuncs import functions, create_signature


@dataclass
class Arg:
    name: str
    ty: types.Type


@dataclass
class Sig:
    name: str
    return_ty: types.Type
    args: tuple[Arg]


@dataclass
class Descriptor:
    py_name: str
    arg_names: tuple[str]
    py_sig: typing.Signature
    c_sig: typing.Signature

    def __str__(self) -> str:
        py_args = ", ".join(f"{name}: {ty}" for ty, name in zip(self.py_sig.args, self.arg_names))
        c_args = ", ".join(f"{ty} {name}" for ty, name in zip(self.c_sig.args, self.arg_names))
        c_name = f"__nv_{self.py_name}"
        c_ret = "void" if self.c_sig.return_type is types.none else str(self.c_sig.return_type)
        c_api = f"{c_ret} {c_name}({c_args});"

        match self.py_sig.return_type:
            case types.UniTuple() as ut:
                py_ret = f"UniTuple({ut.dtype}, {ut.count})"
            case types.Tuple() as t:
                ty_str = ", ".join(str(ty) for ty in t.types)
                py_ret = f"Tuple([{ty_str}])"
            case _:
                py_ret = str(self.py_sig.return_type)

        return f'''
def {self.py_name}({py_args}) -> {py_ret}:
    """
    See https://docs.nvidia.com/cuda/libdevice-users-guide/{c_name}.html

    CAPI: {c_api}
    """
'''


def libdevice_descriptors():
    for c_name, (retty, args) in functions.items():
        py_api = create_signature(retty, args)
        c_api_args = tuple(types.CPointer(arg.ty) if arg.is_ptr else arg.ty for arg in args)
        py_name = c_name[5:]
        arg_names = tuple(arg.name.replace("in", "in_") for arg in args)

        yield Descriptor(
            py_name=py_name,
            arg_names=arg_names,
            py_sig=py_api,
            c_sig=retty(*c_api_args),
        )



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import ctypes
import os
import platform

from numba_cuda_mlir._threading import _LockedCounter

if platform.machine() in ("ARM64", "AMD64"):
    NVPTX64_DATALAYOUT = "e-p:64:64:64-p6:32:32:32-i1:8:8-i8:8:8-i16:16:16-i32:32:32-i64:64:64-i128:128:128-f32:32:32-f64:64:64-f128:128:128-v16:16:16-v32:32:32-v64:64:64-v128:128:128-n16:32:64-a:8:8"
else:
    NVPTX64_DATALAYOUT = "e-i64:64-i128:128-v16:16-v32:32-n16:32:64-S128"
NVPTX64_TRIPLE = "nvptx64-nvidia-cuda"


def _find_mlir_capi_lib():
    import numba_cuda_mlir._mlir._mlir_libs as libs

    name = "MLIRPythonCAPI.dll" if os.name == "nt" else "libMLIRPythonCAPI.so"
    return os.path.join(os.path.dirname(libs.__file__), name)


def _find_llvm_c_lib():
    import numba_cuda_mlir._mlir._mlir_libs as libs

    d = os.path.dirname(libs.__file__)
    if os.name == "nt":
        return None
    return os.path.join(d, "libMLIRPythonCAPI.so")


def _find_modern_to_nvvm_bridge_lib():
    import numba_cuda_mlir._mlir._mlir_libs as libs

    d = os.path.dirname(libs.__file__)
    if os.name == "nt":
        name = "MLIRModernToNVVM.dll"
    elif os.uname().sysname == "Darwin":
        name = "libMLIRModernToNVVM.dylib"
    else:
        name = "libMLIRModernToNVVM.so"
    return os.path.join(d, name)


MLIR_CAPI_LIB_PATH = _find_mlir_capi_lib()
LLVM_C_LIB_PATH = _find_llvm_c_lib()
MODERN_TO_NVVM_BRIDGE_LIB_PATH = _find_modern_to_nvvm_bridge_lib()

_capi = None
_modern_to_nvvm_bridge = None
_modern_bridge_dump_counter = _LockedCounter()


def _load_capi_library(path: str, label: str):
    if not os.path.isfile(path):
        raise FileNotFoundError(f"{label} library not found at {path}")
    return ctypes.CDLL(path)


def _get_capi():
    global _capi
    if _capi is not None:
        return _capi
    if os.name == "nt":
        raise RuntimeError(
            "direct MLIR-to-LLVM pointer translation is not supported on Windows; "
            "use the MLIRModernToNVVM bridge"
        )
    _capi = _load_capi_library(MLIR_CAPI_LIB_PATH, "MLIR/LLVM Python C API")
    _capi.LLVMContextCreate.restype = ctypes.c_void_p
    _capi.LLVMContextCreate.argtypes = []
    _capi.mlirTranslateModuleToLLVMIR.restype = ctypes.c_void_p
    _capi.mlirTranslateModuleToLLVMIR.argtypes = [
        ctypes.c_void_p,
        ctypes.c_void_p,
    ]
    _capi.LLVMPrintModuleToString.restype = ctypes.c_void_p
    _capi.LLVMPrintModuleToString.argtypes = [ctypes.c_void_p]
    _capi.LLVMDisposeMessage.restype = None
    _capi.LLVMDisposeMessage.argtypes = [ctypes.c_void_p]
    _capi.LLVMDisposeModule.restype = None
    _capi.LLVMDisposeModule.argtypes = [ctypes.c_void_p]
    _capi.LLVMContextDispose.restype = None
    _capi.LLVMContextDispose.argtypes = [ctypes.c_void_p]
    return _capi


def _get_modern_to_nvvm_bridge():
    global _modern_to_nvvm_bridge
    if _modern_to_nvvm_bridge is not None:
        return _modern_to_nvvm_bridge

    lib = _load_capi_library(MODERN_TO_NVVM_BRIDGE_LIB_PATH, "MLIR modern to NVVM bridge")
    lib.mlir_modern_to_nvvm_translate_for_libnvvm.restype = ctypes.c_int
    lib.mlir_modern_to_nvvm_translate_for_libnvvm.argtypes = [
        ctypes.c_char_p,
        ctypes.c_size_t,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.POINTER(ctypes.c_void_p),
        ctypes.POINTER(ctypes.c_size_t),
        ctypes.POINTER(ctypes.c_void_p),
    ]
    lib.mlir_modern_to_nvvm_free.restype = None
    lib.mlir_modern_to_nvvm_free.argtypes = [ctypes.c_void_p]
    _modern_to_nvvm_bridge = lib
    return lib


def _op_to_raw_ptr(op):
    """Extract the raw C pointer from an MLIR operation's PyCapsule."""
    capsule = op._CAPIPtr
    get_ptr = ctypes.pythonapi.PyCapsule_GetPointer
    get_ptr.restype = ctypes.c_void_p
    get_ptr.argtypes = [ctypes.py_object, ctypes.c_char_p]
    ptr = get_ptr(capsule, b"numba_cuda_mlir._mlir.ir.Operation._CAPIPtr")
    if not ptr:
        raise ValueError(f"failed to extract C pointer from {op!r}")
    return ptr


def _maybe_dump_modern_bridge_mlir(gpu_module_text):
    target = os.environ.get("NUMBA_CUDA_MLIR_DUMP_MODERN_BRIDGE_MLIR")
    if not target:
        return

    if target.lower() in {"1", "true", "yes", "stderr"}:
        import sys

        print(
            f"=============== Modern bridge MLIR ===============\n\n{gpu_module_text}\n",
            file=sys.stderr,
        )
        return

    from pathlib import Path

    path = Path(target)
    if path.exists() and path.is_dir():
        name = f"modern-bridge-{os.getpid()}-{next(_modern_bridge_dump_counter)}.mlir"
        path = path / name
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(gpu_module_text, encoding="utf-8")


def translate_to_llvmir(op):
    """Translate an MLIR gpu.module operation to an LLVMModuleRef.

    Returns (llvm_mod_ptr, llvm_ctx_ptr) as raw integer pointers.
    """
    capi = _get_capi()
    op_ptr = _op_to_raw_ptr(op)
    llvm_ctx = capi.LLVMContextCreate()
    if not llvm_ctx:
        raise RuntimeError("LLVMContextCreate failed")
    try:
        llvm_mod = capi.mlirTranslateModuleToLLVMIR(op_ptr, llvm_ctx)
    except Exception:
        capi.LLVMContextDispose(llvm_ctx)
        raise
    if not llvm_mod:
        capi.LLVMContextDispose(llvm_ctx)
        raise RuntimeError("mlirTranslateModuleToLLVMIR failed")
    return llvm_mod, llvm_ctx


def translate_gpu_module_to_libnvvm_ir(
    gpu_module_text,
    ctk_major,
    ctk_minor,
    nvvm_ir_version,
    dump=False,
    emit_text_ir=False,
    inspect_llvmir=False,
):
    """Translate gpu.module text to libnvvm-compatible LLVM IR bytes."""
    lib = _get_modern_to_nvvm_bridge()
    _maybe_dump_modern_bridge_mlir(gpu_module_text)
    mlir_bytes = gpu_module_text.encode("utf-8")
    out = ctypes.c_void_p()
    out_len = ctypes.c_size_t()
    err_out = ctypes.c_void_p()

    rc = lib.mlir_modern_to_nvvm_translate_for_libnvvm(
        mlir_bytes,
        len(mlir_bytes),
        ctk_major,
        ctk_minor,
        nvvm_ir_version[0],
        nvvm_ir_version[1],
        nvvm_ir_version[2],
        nvvm_ir_version[3],
        int(bool(dump)),
        int(bool(emit_text_ir)),
        int(bool(inspect_llvmir)),
        ctypes.byref(out),
        ctypes.byref(out_len),
        ctypes.byref(err_out),
    )

    if rc != 0:
        msg = (
            ctypes.string_at(err_out).decode("utf-8", errors="replace")
            if err_out.value
            else "unknown error"
        )
        if err_out.value:
            lib.mlir_modern_to_nvvm_free(err_out)
        raise RuntimeError(f"MLIR modern to NVVM bridge failed: {msg}")

    try:
        return ctypes.string_at(out, out_len.value)
    finally:
        if out.value:
            lib.mlir_modern_to_nvvm_free(out)


def dump_llvmir(llvm_mod_ptr):
    """Return the LLVM IR text for an LLVMModuleRef (for debugging)."""
    capi = _get_capi()
    raw = capi.LLVMPrintModuleToString(llvm_mod_ptr)
    if not raw:
        raise RuntimeError("LLVMPrintModuleToString failed")
    result = ctypes.string_at(raw).decode("utf-8")
    capi.LLVMDisposeMessage(raw)
    return result


def translate_to_llvmir_text(op):
    """Translate an MLIR gpu.module operation to readable LLVM IR text."""
    llvm_mod, llvm_ctx = translate_to_llvmir(op)
    capi = _get_capi()
    try:
        return dump_llvmir(llvm_mod)
    finally:
        capi.LLVMDisposeModule(llvm_mod)
        capi.LLVMContextDispose(llvm_ctx)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

try:
    from pygments.styles.default import DefaultStyle
except ImportError:
    msg = "Please install pygments to see highlighted dumps"
    raise ImportError(msg)

import numba_cuda_mlir.numba_cuda.config
from pygments.styles.manni import ManniStyle
from pygments.styles.monokai import MonokaiStyle
from pygments.styles.native import NativeStyle

from pygments.token import Name

from pygments.style import Style


def by_colorscheme():
    """
    Get appropriate style for highlighting according to
    NUMBA_COLOR_SCHEME setting
    """
    styles = DefaultStyle.styles.copy()
    styles.update(
        {
            Name.Variable: "#888888",
        }
    )
    custom_default = type("CustomDefaultStyle", (Style,), {"styles": styles})

    style_map = {
        "no_color": custom_default,
        "dark_bg": MonokaiStyle,
        "light_bg": ManniStyle,
        "blue_bg": NativeStyle,
        "jupyter_nb": DefaultStyle,
    }

    return style_map[numba.cuda.config.COLOR_SCHEME]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import ctypes
import operator
from numba_cuda_mlir.lowering_utilities.type_conversions import to_numba_type
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AttributeTemplate,
    ConcreteTemplate,
    AbstractTemplate,
    Registry,
    bound_function,
    signature,
)
from numba_cuda_mlir import types


class CTypesType(types.Type):
    def __init__(self, ctype):
        self.ctype = ctype
        super().__init__(f"CTypesType({ctype.__name__})")

    def __repr__(self):
        return f"CTypesType({self.ctype})"
